/*
 * flash.c
 *
 * flash handler
 *
 * Copyright (C) 2004-2008 Atmark Techno, Inc.	All Rights Reserved.
 *
 */

#include "memmap.h"
#include "flash.h"
#include "medium.h"

#ifndef SPI_REGISTER_BASEADDR

#define FLASH_SECTOR_SIZE (0x00010000)

#define PRINT_CHECKSUM

#define SYSTEM_GPIO_BASE (SYSTEM_REGISTER_BASEADDR)
#define SYSTEM_GPIO(v)	 (*(volatile unsigned long *)(SYSTEM_GPIO_BASE) = (v))
#define REBOOT		 (1)

typedef unsigned short __u16;

#define ___swab16(x)							\
	({								\
		__u16 __x = (x);					\
		((__u16)((((__u16)(__x) & (__u16)0x00ffU) << 8) |	\
			 (((__u16)(__x) & (__u16)0xff00U) >> 8) ));	\
	})

#define flash_read(addr)						\
	({ unsigned short __v = (*(volatile unsigned short *)(addr)); __v; })
#define flash_write_cmd(addr, b)					\
	(void)((*(volatile unsigned short *)(addr)) = (___swab16(b)))
#define flash_write(addr, b)					\
	(void)((*(volatile unsigned short *)(addr)) = ((b)))
#define flash_read_cmd(base, offset)						\
	({ 									\
		unsigned short __v =						\
			(*(volatile unsigned short *)(base + (offset<<1)));	\
 		___swab16(__v);							\
	})

static void busy_delay(int time)
{
	volatile int i; 

	for (i = 0; i < time; i++) {}
}

#ifdef PRINT_CHECKSUM
static void print_sum(unsigned long start, unsigned long size, char *lavel)
{
	volatile unsigned short *addr = (volatile unsigned short *)start;
	unsigned short temp;
	unsigned long sum = 0;

	for (; size >= 2; size -= 2) {
		temp = *addr++;
		sum += (temp >> 8) + (temp & 0xff);
	}

	if (size)
		sum += *(volatile unsigned char *)addr;

	myprint(lavel);
	myprint(" checksum: ");
	put_char_hex(sum);
}
#endif

int flash_get_size(void)
{
	int ret;

	flash_write_cmd(FLASH_START, 0xf0);
	flash_write_cmd(FLASH_START + 0xaa, 0x98);
	if (flash_read_cmd(FLASH_START, 0x10) == 'Q' &&
	    flash_read_cmd(FLASH_START, 0x11) == 'R' &&
	    flash_read_cmd(FLASH_START, 0x12) == 'Y')
		ret = flash_read_cmd(FLASH_START, 0x27);
	else
		ret = -1;

	flash_write_cmd(FLASH_START, 0xf0);

	return ret;
}

void flash_copy_from_dram(unsigned long start_addr, unsigned long last_addr)
{
	unsigned long offset_start, offset_last, offset;

	/* Erase Flash */
	offset_start = ((start_addr) & 0x7FFFFE);
	offset_last  = ((last_addr) & 0x7FFFFE);
	myprint("\r\nErasing Flash ...\r\n");
	flash_erase(offset_start, offset_last - offset_start + 2);

	/* Copy SDRAM to Flash */
	offset_start = ((start_addr) & 0x7FFFFE);
	offset_last  = ((last_addr + 0x2) & 0x7FFFFE);
	myprint("Programming Flash ...\r\n");

	/* Program Flash */
	for (offset = offset_start; offset < offset_last; offset += 0x2)
		flash_program(offset, *(unsigned short *)(SDRAM_BASEADDR + offset));

	myprint("done.\r\n");
#ifdef PRINT_CHECKSUM
	unsigned long start = SDRAM_BASEADDR + (start_addr & 0x7FFFFF);
	unsigned long size = (last_addr & 0x7FFFFF) - (start_addr & 0x7FFFFF) + 1;

	print_sum(start, size, "Original");

	start = FLASH_START + (start_addr & 0x7FFFFF);

	print_sum(start, size, " Program");
#endif

	myprint("Reboot.\r\n");
	busy_delay(1000000);
	SYSTEM_GPIO(REBOOT);
}

void flash_copy_to_dram(unsigned int offset)
{
	int i;
	volatile unsigned short *dst = (unsigned short *)BOOTLOADER_LOAD_ADDR;

	for (i = 0; i < BOOTLOADER_SIZE; i += 2)
		*dst++ = flash_read(FLASH_START + offset + i);

	return;
}

void flash_program(unsigned int offset, unsigned short val)
{
	flash_write_cmd(FLASH_START + 0xaaa, 0xaa);
	flash_write_cmd(FLASH_START + 0x554, 0x55);
	flash_write_cmd(FLASH_START + 0xaaa, 0xa0);
	flash_write(FLASH_START + offset, val);

	while (flash_read(FLASH_START + offset) != val);
}

void flash_erase(unsigned int offset, unsigned int size)
{
	unsigned long i;

	flash_write_cmd(FLASH_START + 0xaaa, 0xaa);
	flash_write_cmd(FLASH_START + 0x554, 0x55);
	flash_write_cmd(FLASH_START + 0xaaa, 0x80);
	flash_write_cmd(FLASH_START + 0xaaa, 0xaa);
	flash_write_cmd(FLASH_START + 0x554, 0x55);

	for (i = 0; i < size; i += FLASH_SECTOR_SIZE)
		flash_write_cmd(FLASH_START + offset + i, 0x30);

	while (flash_read(FLASH_START + offset + i - FLASH_SECTOR_SIZE) != 0xffff);

	return;
}

#endif /* !SPI_REGISTER_BASEADDR */
