#############################################################################
##
## Copyright (c) 2007 Xilinx, Inc. All Rights Reserved.
##
## mpmc_v2_1_0.tcl
##
#############################################################################


#***--------------------------------***-----------------------------------***
#
#                               IPLEVEL_DRC_PROC
#
#***--------------------------------***-----------------------------------***


proc check_iplevel_settings {mhsinst} {

    check_partno        $mhsinst
    check_port_validity $mhsinst

    ## Run the arbitration DRC only for custom algorithm 
    set arb_algo [xget_hw_parameter_value $mhsinst "C_ARB0_ALGO"]
    if {[string compare -nocase $arb_algo "CUSTOM"] == 0} {
        check_arb0_slot   $mhsinst
    }

    check_wr_fifo            $mhsinst
    check_fifo_type          $mhsinst
    check_sdma_fifo_type     $mhsinst
    check_fifo_pipeline      $mhsinst 

}

proc check_syslevel_settings {mhsinst} {

    check_splb_native_dwidth $mhsinst
    check_splb_subtype       $mhsinst
    check_splb_overlap       $mhsinst
}

#------------------------------------------
# check C_MEM_PARTNO 
# - The value must be set
#------------------------------------------
proc check_partno { mhsinst } {

    global array_partno_param

    set partno  [string toupper [xget_hw_parameter_value  $mhsinst "C_MEM_PARTNO"]]

    if { [string compare -nocase $partno "CUSTOM"] == 0 || [string compare -nocase $partno "Select a part"] == 0} {

        return

    } elseif {[string compare -nocase $partno "NONE"] == 0 } {

        error "\n The parameter C_MEM_PARTNO must be set.\n" "" "mdt_error" 

    } else {

        set list   [array get array_partno_param $partno,*]

        if {[string length $list] == 0} {

            error "\n The parameter C_MEM_PARTNO=$partno is not found in the memory database.\n" "" "mdt_error" 

        }

    } 

}


#------------------------------------------
# check if multiple bus interfaces are 
# defined, but some are invalid
#------------------------------------------
proc check_port_validity { mhsinst } {

    # check the connectivity for bus interfaces
    # XCLx, SPLBx, SDMA_CTRLx, SDMA_LLx, MPMC_PIMx, PPC440MCx, VFBCx

    for {set x 0} {$x < 8} {incr x 1} {

        check_busif_connectivity $mhsinst "XCL$x"       $x
        check_busif_connectivity $mhsinst "SPLB$x"      $x
        check_busif_connectivity $mhsinst "SDMA_CTRL$x" $x
        check_busif_connectivity $mhsinst "SDMA_LL$x"   $x
        check_busif_connectivity $mhsinst "MPMC_PIM$x"  $x
        check_busif_connectivity $mhsinst "PPC440MC$x"  $x
        check_busif_connectivity $mhsinst "VFBC$x"      $x
    }

}

proc check_busif_connectivity {mhsinst busif port} {

    set connector [xget_hw_busif_value $mhsinst $busif]

    if {[llength $connector] == 0 } {

        return
    }

    # convert BASETYPE value (0-6) to string 
    set i_btype [xget_hw_parameter_value $mhsinst "C_PIM${port}_BASETYPE"]
    set s_btype [lindex {INACTIVE XCL PLB SDMA MPMC_PIM PPC440MC VFBC} $i_btype]
    set portno  [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]

    if { $port >= $portno || [string match -nocase *${s_btype}* $busif] == 0} {

        set instname [xget_hw_parameter_value $mhsinst "INSTANCE"]
        set ipname   [xget_hw_option_value    $mhsinst "IPNAME"]
        puts  "\nWARNING:  $instname ($ipname) - \n      The bus interface $busif is invalid.\n"

    }

}

#------------------------------------------
# check C_ARB0_SLOTx, x = 0:C_ARB0_NUM_SLOT 
# - the number of digit of the value 
#   must be = C_NUM_PORTS
# - each digit must be <= C_NUM_PORTS - 1
#------------------------------------------
proc check_arb0_slot { mhsinst } {

    set slotno   [xget_hw_parameter_value $mhsinst "C_ARB0_NUM_SLOTS"]
    set portno   [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]

    for {set x 0} {$x < $slotno} {incr x 1} {

        set value [xget_hw_parameter_value $mhsinst "C_ARB0_SLOT$x"]

        # check the number of digit of the value
        if { [string length $value] != $portno } {

            error "\n Invalid parameter C_ARB0_SLOT$x = $value.  The value must be $portno digits.\n" "" "mdt_error"


        }

        # check each digit must be <= C_NUM_PORTS - 1
        #       no duplicate digit
        set listDigit ""

        for {set i 0} {$i < [string length $value]} {incr i 1} {

            set digit [string index $value $i]
            if {$digit >= $portno} {

                error "\n Invalid parameter C_ARB0_SLOT$x = $value.  Each digit must be less than or equal to [expr $portno - 1]\n" "" "mdt_error" 
        
            }

            if {[lsearch $listDigit $digit] == -1} {

                lappend listDigit $digit

            } else {

                error "\n Invalid parameter C_ARB0_SLOT$x = $value.  Each digit must be unique\n" "" "mdt_error" 
            
            }
 
        }

    }

}


#------------------------------------------
# for ports x < C_NUM_PORTS
# if C_PIMx_SUBTYPE != IXCL && C_XCLx_WRITEXFER > 0 AND C_PIMx_BASTYPE == 1
#    AND C_PIx_WR_FIFO_TYPE = DISABLED
# then error out
#------------------------------------------
proc check_wr_fifo { mhsinst } {

    set portno [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]

    for {set x 0} {$x < $portno} {incr x 1} {

        set xfer [xget_hw_parameter_value $mhsinst "C_XCL${x}_WRITEXFER"]
        set wr   [xget_hw_parameter_value $mhsinst "C_PI${x}_WR_FIFO_TYPE"]
        set basetype  [xget_hw_parameter_value $mhsinst "C_PIM${x}_BASETYPE"]
        set subtype  [xget_hw_parameter_value $mhsinst "C_PIM${x}_SUBTYPE"]

        if {[string compare -nocase $subtype "IXCL"] != 0 && $basetype == 1 && $xfer > 0 && [string compare -nocase $wr "DISABLED"] == 0} {

            error "\n XCL port $x cannot be configured to be Read-Only (C_PI${x}_WR_FIFO_TYPE = DISABLED) when C_XCL${x}_WRITEXFER > 0.\n" "" "mdt_error" 

        }
    }
}


#------------------------------------------
# for ports x < C_NUM_PORTS
# if C_PIx_RD_FIFO_TYPE = DISABLED AND 
#    C_PIx_WR_FIFO_TYPE = DISABLED
# then error out
#------------------------------------------
proc check_fifo_type { mhsinst } {

    set portno [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]

    for {set x 0} {$x < $portno} {incr x 1} {

        set rd [xget_hw_parameter_value $mhsinst "C_PI${x}_RD_FIFO_TYPE"]
        set wr [xget_hw_parameter_value $mhsinst "C_PI${x}_WR_FIFO_TYPE"]
        
        if {[string compare -nocase $rd "DISABLED"] == 0 && [string compare -nocase $wr "DISABLED"] == 0} {

            error "\n Port $x cannot be configured to be both Read-Only and Write-Only.\n" "" "mdt_error" 
        
        }
    }
}

#------------------------------------------
# check port x < C_NUM_PORTS and C_PIMx_BASETYPE = 3 (sdma)
# if C_PIx_RD_FIFO_TYPE = DISABLED OR
#    C_PIx_WR_FIFO_TYPE = DISABLED
# then error out
#------------------------------------------
proc check_sdma_fifo_type { mhsinst } {

    set portno [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]

    for {set x 0} {$x < $portno} {incr x 1} {

        set btype [xget_hw_parameter_value $mhsinst "C_PIM${x}_BASETYPE"]

        if {$btype != 3} {

            continue
        }

        set rd    [xget_hw_parameter_value $mhsinst "C_PI${x}_RD_FIFO_TYPE"]
        set wr    [xget_hw_parameter_value $mhsinst "C_PI${x}_WR_FIFO_TYPE"]
        
        if {[string compare -nocase $rd "DISABLED"] == 0 || [string compare -nocase $wr "DISABLED"] == 0} {

            error "\n SDMA Port $x cannot be configured to be Read-Only or Write-Only.  Please ensure that writes and reads FIFO types are SRL or BRAM.\n" "" "mdt_error" 
        
        }
    }

}


#------------------------------------------
# for ports x < C_NUM_PORTS
# C_PIx_RD_FIFO_MEM_PIPELINE must be all the same
# C_PIx_WR_FIFO_MEM_PIPELINE must be all the same
# If the FIFOs are DISABLED or if the PORT is INACTIVE, then these are not to be considered
#------------------------------------------
proc check_fifo_pipeline { mhsinst } {

    set portno [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]
    set rd_pipeline_mode ""
    set wr_pipeline_mode ""

    for {set x 0} {$x < $portno} {incr x 1} {

        set rd_x_fifo_type [xget_hw_parameter_value $mhsinst "C_PI${x}_RD_FIFO_TYPE"]
        set wr_x_fifo_type [xget_hw_parameter_value $mhsinst "C_PI${x}_WR_FIFO_TYPE"]

        set rd [xget_hw_parameter_value $mhsinst "C_PI${x}_RD_FIFO_MEM_PIPELINE"]
        set wr [xget_hw_parameter_value $mhsinst "C_PI${x}_WR_FIFO_MEM_PIPELINE"]

        set i_btype [xget_hw_parameter_value $mhsinst "C_PIM${x}_BASETYPE"]
        set s_btype [lindex {INACTIVE XCL PLB SDMA MPMC_PIM PPC440MC VFBC} $i_btype]

        if { [string compare -nocase $s_btype "INACTIVE"] != 0 } {  ## Port is valid 

            ## Read Fifo
            if { [ string compare -nocase $rd_x_fifo_type "DISABLED" ] != 0 } { ## Port is not disabled
                if { [string compare -nocase $rd_pipeline_mode ""] != 0 } { ## Check if the initial RD fifo mode has been set
                    if { $rd_pipeline_mode != $rd } { ## Compare the pipeline set to prior ports. Fail if not matching
                        error "\nRead Memory Pipeline setting for all MPMC ports should be same. Check the memory pipeline settings in the Data Configuration section of the Advanced tab of Memory configuration\n" "" "mdt_error"
                    }
                } else { ## This is the first valid read fifo
                    set rd_pipeline_mode $rd
                }
            }

            ## Write Fifo
            if { [ string compare -nocase $wr_x_fifo_type "DISABLED" ] != 0 } { ## Port is not disabled
                if { [string compare -nocase $wr_pipeline_mode ""] != 0 } { ## Check if the initial WR fifo mode has been set
                    if { $wr != $wr_pipeline_mode } { ## Compare the pipeline set to prior ports. Fail if not matching
                         error "\nWrite Memory Pipeline setting for all MPMC ports should be same. Check the memory pipeline settings in the Data Configuration section of the Advanced tab of Memory configuration.\n" "" "mdt_error"
                    }
                } else { ## This is the first valid write fifo
                    set wr_pipeline_mode $wr
                }
            }
        }
    }
}

#------------------------------------------
# if C_FAMILY = virtex5 AND
#    C_MEM_TYPE = DDR2 AND
#    C_USE_STATIC_PHY = 0 AND
#    C_DDR2_DQSN_ENABLE = 0
# then error out
# or  
# if C_FAMILY = spartan3 AND
#    C_MEM_TYPE = DDR2 AND
#    C_DDR2_DQSN_ENABLE = 1
# then error out
# or  
# if C_FAMILY = aspartan3 AND
#    C_MEM_TYPE = DDR2 AND
#    C_DDR2_DQSN_ENABLE = 1
# then error out
#------------------------------------------
proc check_ddr2_dqsn_enable { param_handle } {

    set mhsinst     [xget_hw_parent_handle    $param_handle]

    set family      [xget_hw_parameter_value $mhsinst "C_FAMILY"]
    set mem_type    [xget_hw_parameter_value $mhsinst "C_MEM_TYPE"]
    set static_phy  [xget_hw_parameter_value $mhsinst "C_USE_STATIC_PHY"]
    set ddr2_dqsn [xget_hw_parameter_value $mhsinst "C_DDR2_DQSN_ENABLE"]

    if { [ string compare -nocase $family "virtex5" ] == 0 && [ string compare -nocase $mem_type "DDR2" ] == 0 && $static_phy == 0 && $ddr2_dqsn == 0 } {

        set instname [xget_hw_parameter_value $mhsinst "INSTANCE"]

        error "\n $instname\:\:C_DDR2_DQSN_ENABLE must be set to 1 when using DDR2 on Virtex-5 and not using the static phy.\n" "" "mdt_error" 

    }
    if { [ string compare -nocase $family "spartan3" ] == 0 && [ string compare -nocase $mem_type "DDR2" ] == 0 && $ddr2_dqsn == 1 } {

        set instname [xget_hw_parameter_value $mhsinst "INSTANCE"]

        error "This parameter must be set to 0 when using DDR2 on Spartan-3 since this family does not support differential IO.\n" "" "mdt_error" 

    }
    if { [ string compare -nocase $family "aspartan3" ] == 0 && [ string compare -nocase $mem_type "DDR2" ] == 0 && $ddr2_dqsn == 1 } {

        set instname [xget_hw_parameter_value $mhsinst "INSTANCE"]

        error "This parameter must be set to 0 when using DDR2 on Spartan-3 since this family does not support differential IO.\n" "" "mdt_error" 

    }


}

#------------------------------------------
# if PARAM & C_MEM_NUM_RANKS*C_MEM_NUM_DIMMS != 0 then error
#------------------------------------------
proc check_ranks_x_dimms_multiple { param_handle } {

    set mhsinst     [xget_hw_parent_handle    $param_handle]
    set ranks       [xget_hw_parameter_value $mhsinst "C_MEM_NUM_RANKS"]
    set dimms       [xget_hw_parameter_value $mhsinst "C_MEM_NUM_DIMMS"]
    set my_param    [xget_value $param_handle]

    if { [expr ${my_param} % (${ranks} * ${dimms}) ] != 0 } {

        set instname [xget_hw_parameter_value $mhsinst "INSTANCE"]

        error "This parameter must be set to a multiple of C_MEM_NUM_RANKS*C_MEM_NUM_DIMMS.\n" "" "mdt_error" 

    }
}

#------------------------------------------
# if C_PIMx_BASETYPE == 2 (PLB) AND
#    C_SPLBx_NATIVE_DWIDTH = 32 AND 
#    C_SPLBx_DWIDTH        = 128
# then error out
#------------------------------------------
proc check_splb_native_dwidth { mhsinst } {

    set portno [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]

    for {set x 0} {$x < $portno} {incr x 1} {

        set native [xget_hw_parameter_value $mhsinst "C_SPLB${x}_NATIVE_DWIDTH"]
        set dwidth [xget_hw_parameter_value $mhsinst "C_SPLB${x}_DWIDTH"]
        set basetype [xget_hw_parameter_value $mhsinst "C_PIM${x}_BASETYPE"]
        if {$basetype == 2 && $native == 32 && [expr {$dwidth == 128 || $dwidth == 64 }]} {

            set instname [xget_hw_parameter_value $mhsinst "INSTANCE"]

            error "\n $instname\:\:C_SPLB$x has native width set to 32. The bus connected to the port has width of $dwidth. A 32 bit MPMC port can only be connected to a 32 bit wide bus.\n" "" "mdt_error" 

        }
    }

}


#-------------------------------------------
# - if C_SPLBx_SUBTYPE is DPLB or IPLB then 
#     if C_SPLB_NATIVE_DWIDTH != 64 then error out
#     if C_SPLB_DWIDTH != 64 then error out
#     if C_SPLB_P2P != 1 then error out 
#     if C_SPLB_SMALLEST_MASTER != 64 then error out
#
#-------------------------------------------
proc check_splb_subtype { mhsinst } {

    set portno [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]

    for {set x 0} {${x} < $portno} {incr x 1} {
        
        set basetype   [xget_hw_parameter_value $mhsinst "C_PIM${x}_BASETYPE"]

        if {$basetype == 2} {

            set subtype [xget_hw_parameter_value $mhsinst "C_PIM${x}_SUBTYPE"]

            if { [string compare -nocase "IPLB" $subtype ] == 0 || [string compare -nocase "DPLB" $subtype ] == 0} {

                set native_dwidth   [xget_hw_parameter_value $mhsinst "C_SPLB${x}_NATIVE_DWIDTH"]
                set dwidth          [xget_hw_parameter_value $mhsinst "C_SPLB${x}_DWIDTH"]
                set p2p             [xget_hw_parameter_value $mhsinst "C_SPLB${x}_P2P"]
                set smallest_master [xget_hw_parameter_value $mhsinst "C_SPLB${x}_SMALLEST_MASTER"]
                set instname        [xget_hw_parameter_value $mhsinst "INSTANCE"]

                if {$native_dwidth != 64} {

                    error "\n $instname\:\:SPLB${x} is set to DPLB or IPLB subtype but C_SPLB${x}_NATIVE_DWIDTH is not set to 64.  This combination is invalid.\n" "" "mdt_error" 
                }
                if {$dwidth != 64} {

                    error "\n $instname\:\:SPLB${x} is set to DPLB or IPLB subtype but C_SPLB${x}_DWIDTH is not set to 64.  This combination is invalid.\n" "" "mdt_error" 
                }
                if {$p2p != 1} {

                    error "\n $instname\:\:SPLB${x} is set to DPLB or IPLB subtype but C_SPLB${x}_P2P is not set to 1.  This combination is invalid.\n" "" "mdt_error" 
                }
                if {$smallest_master != 64} {

                    error "\n $instname\:\:SPLB${x} is set to DPLB or IPLB subtype but C_SPLB${x}_SMALLEST_MASTER is not set to 64.  This combination is invalid.\n" "" "mdt_error" 
                }
            }
        }
    }
}


#------------------------------------------
# if C_ALL_PIM_SHARE_ADDRESSES
# for all valid ports
# if C_PIMx_BASETYPE == 2 (PLB) AND
#    C_PIMy_BASETYPE == 2, where x != y
# then error out
#------------------------------------------
proc check_splb_overlap { mhsinst } {

    set portno [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]
    set shared [xget_hw_parameter_value $mhsinst "C_ALL_PIMS_SHARE_ADDRESSES"]

    if {$shared == 0} {
        return
    }

    set x_max [expr $portno - 1]

    for {set x 0} {$x < $x_max} {incr x 1} {

        set y_start [expr $x + 1]

        for {set y $y_start} {$y < $portno} {incr y 1} {

            set pimx_basetype [xget_hw_parameter_value $mhsinst "C_PIM${x}_BASETYPE"]
            set pimy_basetype [xget_hw_parameter_value $mhsinst "C_PIM${y}_BASETYPE"]

            if { $pimx_basetype == 2 && $pimy_basetype == 2 && (${x} != ${y}) } { 

                set splbx_connector [xget_hw_busif_value $mhsinst "SPLB${x}"]
                set splby_connector [xget_hw_busif_value $mhsinst "SPLB${y}"]

                if {[string compare -nocase $splbx_connector $splby_connector] == 0} {

                    set instname [xget_hw_parameter_value $mhsinst "INSTANCE"]

                    error "\n $instname\:\:SPLB${x} and $instname\:\:SPLB${y} are connected to the same PLB bus and have overlapping addresses.\n" "" "mdt_error" 

                }
            }
        }
    }

}


proc gui_drc_pim_type { mhsinst } {
    set family      [xget_hw_parameter_value $mhsinst "C_FAMILY"]
    set subfamily      [xget_hw_parameter_value $mhsinst "C_SUBFAMILY"]

    for {set i 0} {$i < 8} {incr i 1} {

        set param_handle [xget_hw_parameter_handle $mhsinst "C_PIM${i}_BASETYPE"]
        set my_param [xget_hw_value $param_handle ]
        if { [ string compare -nocase $family "virtex2p" ] == 0 && \
                [ string compare -nocase $my_param "6" ] == 0 } {
            xset_hw_parameter_value $param_handle "0"
            error "Port type VFBC is not supported on ${family}${subfamily}"
        } elseif { [ string compare -nocase $family "virtex2p" ] == 0 && \
                [ string compare -nocase $my_param "5" ] == 0 } {
            xset_hw_parameter_value $param_handle "0"
            error "Port type PPC440MC is not supported on ${family}${subfamily}"
        } elseif { ([ string compare -nocase $family "virtex5" ] != 0 || \
                   [ string compare -nocase $subfamily "fx" ] != 0) && \
                   [ string compare -nocase $my_param "5" ] == 0 } {
            xset_hw_parameter_value $param_handle 0
            error "Port type PPC440MC is not supported on ${family}${subfamily}"
        }
    } 

    # C_NUM_PORTS
    update_portno        $mhsinst
    mpmc_update_arbitration  $mhsinst
}
#***--------------------------------***------------------------------------***
#
#                        IPLEVEL_UPDATE_VALUE_PROC
#
#***--------------------------------***------------------------------------***

#-------------------------------------------
# given C_MEM_PARTNO
# update a single memory parameter 
#-------------------------------------------
proc iplevel_update_mem_parameter { param_handle } {

    global array_partno_param

    set mhsinst    [xget_hw_parent_handle    $param_handle]
    init_memory    $mhsinst
    set partno     [string toupper [xget_hw_parameter_value  $mhsinst "C_MEM_PARTNO"]]
    set paramname  [string toupper [xget_hw_name             $param_handle]]
    set paramvalue [xget_hw_value            $param_handle]

    if { [string compare -nocase $partno "CUSTOM"] == 0 } {

        return $paramvalue

    } elseif {[string compare -nocase $partno "NONE"] == 0 } {

        return $paramvalue 

    } else {

        set key_value  [array get array_partno_param $partno,$paramname]
        set value      [lindex $key_value 1]

        if {[string length $value] == 0} {

            return  $paramvalue

        } else {

            return  $value

        }
    }
}


#-------------------------------------------
# update C_PIX_DATA_WIDTH_MAX
# if an C_PIMx_DATA_WIDTH = 64(x=0: C_NUM_PORTS - 1), then 64  
# else                               32
#-------------------------------------------
proc iplevel_update_pix_dwidth { param_handle } {

    set mhsinst     [xget_hw_parent_handle   $param_handle]
    set portno      [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]

    for {set i 0} {$i < $portno} {incr i 1} {

        if { [xget_hw_parameter_value $mhsinst "C_PIM${i}_DATA_WIDTH"] == 64 } {

            return 64

        }

    }

    return 32

}

#-------------------------------------------
# update C_MEM_DM_WIDTH
# if C_MEM_BITS_DATA_PER_DQS == 8 then C_MEM_DATA_WIDTH/8
# else                            then 1
#-------------------------------------------
proc iplevel_update_mem_dm_width { param_handle } {

    set mhsinst     [xget_hw_parent_handle   $param_handle]
    set dwidth      [xget_hw_parameter_value $mhsinst "C_MEM_DATA_WIDTH"]
    set bitsdata    [xget_hw_parameter_value $mhsinst "C_MEM_BITS_DATA_PER_DQS"]

    if {$bitsdata == 8} {

        return [expr $dwidth/8]

    } else {

        return 1

    }

}


#-------------------------------------------
# update C_MEM_DQS_WIDTH
# = C_MEM_DATA_WIDTH/C_MEM_BITS_DATA_PER_DQS
#-------------------------------------------
proc iplevel_update_mem_dqs_width { param_handle } {

    set mhsinst     [xget_hw_parent_handle   $param_handle]
    set dwidth      [xget_hw_parameter_value $mhsinst "C_MEM_DATA_WIDTH"]
    set bitsdata    [xget_hw_parameter_value $mhsinst "C_MEM_BITS_DATA_PER_DQS"]

    return [expr $dwidth/$bitsdata]

}



#-------------------------------------------
# update C_MEM_BITS_DATA_PER_DQS 
# if C_MEM_PART_DATA_WIDTH == 4 then 4
# else                               8
#-------------------------------------------
proc iplevel_update_mem_bits_data { param_handle } {

    set mhsinst     [xget_hw_parent_handle   $param_handle]
    set dwidth      [xget_hw_parameter_value $mhsinst "C_MEM_PART_DATA_WIDTH"]

    if {$dwidth == 4} {

        return 4

    } else {

        return 8

    }

}


#-------------------------------------------
# update C_ECC_DATA_WIDTH 
#-------------------------------------------
proc iplevel_update_ecc_datawidth { param_handle } {

    set mhsinst     [xget_hw_parent_handle   $param_handle]
    set paramName   [xget_hw_name            $param_handle]
    set ecc_support [xget_hw_parameter_value $mhsinst "C_INCLUDE_ECC_SUPPORT"]
    set mem_type    [xget_hw_parameter_value $mhsinst "C_MEM_TYPE"]
    set mem_dwidth  [xget_hw_parameter_value $mhsinst "C_MEM_DATA_WIDTH"]

    if {$ecc_support == 0} {

        return 0

    } else {

        if {[string compare -nocase $mem_type "SDRAM"] == 0} { # SDRAM

            switch $mem_dwidth {
                8       { return 5 }
                16      { return 6 }
                32      { return 7 }
                64      { return 8 }
                default { return 8 }
            }

        } else {  # DDR/DDR2

            switch $mem_dwidth {
                8       { return 8 }
                16      { return 8 }
                32      { return 8 }
                64      { return 8 }
                default { return 8 }
            }
        }
    }

}


#-------------------------------------------
# update C_ECC_DM_WIDTH 
#-------------------------------------------
proc iplevel_update_ecc_dmwidth { param_handle } {

    set mhsinst     [xget_hw_parent_handle   $param_handle]
    set paramName   [xget_hw_name            $param_handle]
    set ecc_support [xget_hw_parameter_value $mhsinst "C_INCLUDE_ECC_SUPPORT"]
    set mem_bits    [xget_hw_parameter_value $mhsinst "C_MEM_BITS_DATA_PER_DQS"]

    if {$ecc_support == 0} {

        return 0

    } else {

        if {$mem_bits == 4} {

            return 0

        } else {

            return 1

        }
    }
}


#-------------------------------------------
# update C_ECC_DQS_WIDTH 
#-------------------------------------------
proc iplevel_update_ecc_dqswidth { param_handle } {

    set mhsinst     [xget_hw_parent_handle   $param_handle]
    set paramName   [xget_hw_name            $param_handle]
    set ecc_support [xget_hw_parameter_value $mhsinst "C_INCLUDE_ECC_SUPPORT"]
    set mem_bits    [xget_hw_parameter_value $mhsinst "C_MEM_BITS_DATA_PER_DQS"]
    set ecc_dwidth  [xget_hw_parameter_value $mhsinst "C_ECC_DATA_WIDTH"]

    if {$ecc_support == 0} {

        return 0

    } else {

        if {$ecc_dwidth > 4 && $mem_bits == 4} {

            return 2

        } else {

            return 1

        }
    }
}


#-------------------------------------------
# update C_MEM_BANKADDR_WIDTH 
# C_MEM_BANKADDR_WIDTH = C_MEM_PART_NUM_BANK_BITS
#-------------------------------------------
proc iplevel_update_bankaddr_width {param_handle}  {

    set mhsinst   [xget_hw_parent_handle   $param_handle]
    set bank_bits [xget_hw_parameter_value $mhsinst "C_MEM_PART_NUM_BANK_BITS"]

    return $bank_bits

}


#-------------------------------------------
# update C_MEM_ADDR_WIDTH 
# C_MEM_ADDR_WIDTH = C_MEM_PART_NUM_ROW_BITS
#-------------------------------------------
proc iplevel_update_addr_width {param_handle}  {

    set mhsinst  [xget_hw_parent_handle   $param_handle]
    set row_bits [xget_hw_parameter_value $mhsinst "C_MEM_PART_NUM_ROW_BITS"]

    return $row_bits

}


#-------------------------------------------
# update C_ARB_BRAM_INIT_xx by converting
#-------------------------------------------
proc iplevel_update_bram_init {param_handle}  {

    set mhsinst    [xget_hw_parent_handle   $param_handle]
    set paramname  [xget_hw_name            $param_handle]
    set paramvalue [xget_hw_value           $param_handle]

    set arb_algo   [xget_hw_parameter_value $mhsinst "C_ARB0_ALGO"]

    if {[string compare -nocase $arb_algo "CUSTOM"] != 0} {
        return $paramvalue 
    }

    set slotno [xget_hw_parameter_value  $mhsinst "C_ARB0_NUM_SLOTS"]
    set no_0 $slotno
    set no_1 0

    if {$slotno > 8} {

        set no_0 8
        set no_1 [expr $slotno - 8]

    }

    # if C_ARB0_NUM_SLOT = 10
    # then C_ARB_BRAM_INIT_00 is generated from C_ARB0_SLOT0 to C_ARB0_SLOT7
    # then C_ARB_BRAM_INIT_01 is generated from C_ARB0_SLOT8 to C_ARB0_SLOT9
    if {[string compare -nocase "C_ARB_BRAM_INIT_00" $paramname] == 0} {

        return [get_bram_addr $param_handle 0 [expr $no_0 - 1]] 

    } elseif {[string compare -nocase "C_ARB_BRAM_INIT_01" $paramname] == 0} {

        return [get_bram_addr $param_handle 8 [expr $no_1 + 7]]

    } else {

        return $paramvalue

    }

}


#-------------------------------------------
# convert C_ARBx_SLOTx to C_ART_BRAM_INIT_xx string
#-------------------------------------------
proc get_bram_addr {param_handle startSlot endSlot} {

    set mhsinst    [xget_hw_parent_handle    $param_handle]

    # the digit number of C_ARB0_SLOTx = C_NUM_PORTS
    # ex. if C_NUM_PORTS = 3 then C_ARB0_SLOT0 = "012" - 3 digits
    set portno     [xget_hw_parameter_value  $mhsinst "C_NUM_PORTS"]

    set slotno     [expr $endSlot - $startSlot + 1]
    set max_slotno       8
    set unused_slot_pads "00000000111111111111111111111111"

    if {$slotno == 0} {

        for {set i $slotno} {$i < $max_slotno } {incr i 1} {

            append addr $unused_slot_pads
        }

        return [format "0b%s" $addr]

    }

    set bram_addr       ""
    set slot_head       "00000000"
    set templatePort    ""
    set list_port       ""
    set used_slot_list  ""

    # update C_ARB_BRAM_INIT_xx, convert from C_ARB0_SLOTx, x = startSlot:endSlot
    for {set x $endSlot} {$x >= $startSlot} {incr x -1} {

        lappend used_slot_list [xget_hw_parameter_value $mhsinst "C_ARB0_SLOT$x"]

    }

    #---------------------
    # fill up unused slots
    #---------------------
    for {set i $slotno} {$i < $max_slotno } {incr i 1} {

        append bram_addr $unused_slot_pads
    }

    #---------------------
    # fill up used slots
    #---------------------
    # 1. fill up unused pads to a used slot
    set unused_pad  ""

    for {set i $portno} {$i < 8 } {incr i 1} {
        append unused_pad "111"
    }

    append slot_head $unused_pad

    # 2. get the used slot
    set list_slot ""

    for {set i 0} {$i < $slotno} {incr i 1} {

        set slot          [lindex $used_slot_list $i]
        set binary_format $slot_head

        for {set m [expr [string length $slot] - 1]} {$m >= 0} {incr m -1} {

            append binary_format [convert_num_to_binary [string index $slot $m] 3]

        }

        lappend list_slot $binary_format
    }

    for {set i 0} {$i < $slotno } {incr i 1} {

        append bram_addr [lindex $list_slot $i]
    }

    return [format "0b%s" $bram_addr]
}


#-------------------------------------------
# convert a number to a binary
# @ value       a number
# @ length      number of digit in binary
#-------------------------------------------
proc convert_num_to_binary {value length} {

    if {$length == 0} {

         error "xconvert_num_to_binary expects lenght to be a positive number" "" "mdt_error"
        return
    }

    set retval ""

    while {$value > 0} {

        set tval [expr $value + 1]
        set val_by_2 [expr $value/2]
        set tval_by_2 [expr $tval/2]
        if {$val_by_2 == $tval_by_2} {
            # means $val is even
            set retval [format "0%s" $retval]
        } else {
            # means $val is odd
            set retval [format "1%s" $retval]
        }
        set value [expr $value/2]
    }

    while {$length > [string length $retval]} {
        set retval [format "0%s" $retval]
    }

    return $retval
}


#***--------------------------------***-----------------------------------***
#
#                            SYSLEVEL_UPDATE_VALUE_PROC
#
#***--------------------------------***-----------------------------------***

#-------------------------------------------
# update C_CTRL_* parameter
#-------------------------------------------
proc syslevel_update_ctrl_parameter { param_handle } {

    global array_ctrl_param

    set mhsinst    [xget_hw_parent_handle    $param_handle]
    init_control   $mhsinst
    set paramname  [xget_hw_name             $param_handle]
    set paramvalue [xget_hw_value            $param_handle]

    set newvalue   [lindex [array get array_ctrl_param $paramname] 1]

    if {[string length $newvalue] == 0} {

        return  $paramvalue

    } else {

        return  $newvalue

    }

}


#-------------------------------------------
# update C_PIMx_SUBTYPE, x = 0:C_NUM_PORTS - 1
#
# - if XCLx is connected
#   set to IXCL  or DXCL  if it's connectoed to MB
#
# - if SPLBx is connected for PPC405 architectures & 
#   if #master connectors = 1 && #slave connectors = 1
#   if SPLBx_NATIVE_DWIDTH == 64 && SPLBx_DATA_WIDTH == 64
#      && SPLBx_SMALLEST_MASTER == 64
#       set to DPLB, if busif = DPLB* -- * can only be 0 or 1. 
#       set to IPLB, if busif = IPLB* -- * can only be 0 or 1.
# - else
#   set to C_PIMx_BASETYPE 
#
# Note that on PPC405 architectures the bus interfaces for PLB are *PLB0 & *PLB1
#-------------------------------------------
proc syslevel_update_pim_subtype { param_handle } {

    set mhsinst    [xget_hw_parent_handle   $param_handle]
    set param_name [xget_hw_name            $param_handle]
    set portno     [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]
    set list_btype {INACTIVE XCL PLB SDMA NPI PPC440MC VFBC}
    set x          [string index $param_name 5]
    set basetype   [xget_hw_parameter_value $mhsinst "C_PIM${x}_BASETYPE"]

    if {$x >= $portno} {
        return [xget_hw_value $param_handle]
    }

    set mhs_handle [xget_hw_parent_handle $mhsinst]

    # check the connectivity for bus XCLx - p2p, target vs. initiator
    if {$basetype == 1} {
        set connector [xget_hw_busif_value $mhsinst "XCL$x"]

        if {[llength $connector] != 0} {
            set busifs [xget_hw_connected_busifs_handle $mhs_handle $connector "initiator"]

                if {[string length $busifs] != 0 } {
                    set busif_parent [xget_hw_parent_handle $busifs]
                    set iptype       [xget_hw_option_value  $busif_parent "IPTYPE"]

                    if { [string compare -nocase "PROCESSOR" $iptype] == 0 } {
                        return [xget_hw_name $busifs]
                }
            }
        }

        return "XCL"
    } elseif {$basetype == 2} {
        # check the connectivity for bus SPLBx
        set connector [xget_hw_busif_value $mhsinst "SPLB$x"]

        if {[llength $connector] != 0} {
            set m_busifs  [xget_hw_connected_busifs_handle $mhs_handle $connector "master"]
            set s_busifs [xget_hw_connected_busifs_handle $mhs_handle $connector "slave"]

            # MASTER connectors number must be 1 
            # SLAVE connectors number must be 1 (MPMC itself)
            if {[llength $m_busifs] == 1 && [llength $s_busifs] == 1} {
                set native_dwidth [xget_hw_parameter_value $mhsinst "C_SPLB${x}_NATIVE_DWIDTH"]
                set dwidth [xget_hw_parameter_value $mhsinst "C_SPLB${x}_DWIDTH"]
                set smallest_master [xget_hw_parameter_value $mhsinst "C_SPLB${x}_SMALLEST_MASTER"]
                set busif_parent [xget_hw_parent_handle $m_busifs]
                set iptype [xget_hw_option_value  $busif_parent "IPTYPE"]

                if {$native_dwidth == 64 && $dwidth == 64 && $smallest_master == 64} { 
                    if { [string compare -nocase "PROCESSOR" $iptype] == 0} {
                        set busif_name [xget_hw_name $m_busifs]

                        if {[string compare -nocase DPLB1 $busif_name] == 0 || [string compare -nocase DPLB0 $busif_name] == 0} {
                            return "DPLB"
                        } elseif {[string compare -nocase IPLB1 $busif_name] == 0 || [string compare -nocase IPLB1 $busif_name] == 0} {
                            return "IPLB"
                        } 
                    }
                }
            }
        }

        return "PLB"

    } else {
        return [lindex $list_btype $basetype]
    }
}


#-------------------------------------------
# update C_WR_TRAINING_PORT = x, when the 
# first port (0-7) found that satisfies 
# - C_PIMx_SUBTYPE != IXCL && IPLB && INACTIVE
# - C_PIMx_BASETYPE == 1 && C_XCLx_WRITEXFER > 0
# - C_PIx_WR_FIFO_TYPE = BRAM or SRL
#-------------------------------------------
proc syslevel_update_training_port { param_handle } {

    set mhsinst [xget_hw_parent_handle   $param_handle]
    set portno  [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]

    for {set x 0} {$x < $portno} {incr x 1} {

        set subtype  [xget_hw_parameter_value $mhsinst "C_PIM${x}_SUBTYPE"]
        set basetype  [xget_hw_parameter_value $mhsinst "C_PIM${x}_BASETYPE"]
        set writexfer [xget_hw_parameter_value $mhsinst "C_XCL${x}_WRITEXFER"]
        if {[string compare -nocase $subtype "IXCL"] != 0 && [string compare -nocase $subtype "IPLB"] != 0 && [string compare -nocase $subtype "INACTIVE"] != 0} {

            if {((($basetype == 1) && ($writexfer > 0)) || ($basetype != 1))} { 

                set fifotype [xget_hw_parameter_value $mhsinst "C_PI${x}_WR_FIFO_TYPE"]

                if {[string compare -nocase $fifotype "BRAM"] == 0 || [string compare -nocase $fifotype "SRL"] == 0} {

                    return $x
                }
            }
        }
    }

    # not found, default to 0
    return 0
}


#-------------------------------------------
# update C_XCLx_WRITEXFER, x = 0:C_NUM_PORTS - 1
# if C_PIMx_SUBTYPE = IXCL 
# then C_XCLx_WRITEXFER = 0
#-------------------------------------------
proc syslevel_update_writexfer { param_handle } {

    set mhsinst    [xget_hw_parent_handle   $param_handle]
    set portno     [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]
    set param_name [xget_hw_name            $param_handle]
    set x          [string index $param_name 5]
    set subtype    [xget_hw_parameter_value $mhsinst "C_PIM${x}_SUBTYPE"]

    if {$x < $portno && [string compare -nocase "IXCL" $subtype] == 0} {

        return 0

    } else {

        return [xget_hw_value $param_handle]
    }
}  


#***--------------------------------***-----------------------------------***
#
#                            SYSLEVEL_DRC_PROC
#
#***--------------------------------***-----------------------------------***

#-------------------------------------------
# check C_WR_TRAINING_PORT
# check that there is at least 1 writable port.
# Cannot have only IXCL and IPLB PIMs
# Port has to be writeable, C_XCL0_WRITEXFER == 0 is invalid or if 
# Wr_FIFO_TYPE is disabled
#-------------------------------------------
proc check_training_port {param_handle} {

    set mhsinst [xget_hw_parent_handle   $param_handle]

    # get the port number from C_WR_TRAINING_PORT
    set x       [xget_hw_value $param_handle]

    # check C_PIMx_SUBTYPE != IXCL or IPLB AND
    # C_PIx_WR_FIFO_TYPE    = BRAM or SRL
    set subtype  [xget_hw_parameter_value $mhsinst "C_PIM${x}_SUBTYPE"]
    set fifotype [xget_hw_parameter_value $mhsinst "C_PI${x}_WR_FIFO_TYPE"]
    set basetype  [xget_hw_parameter_value $mhsinst "C_PIM${x}_BASETYPE"]
    set writexfer [xget_hw_parameter_value $mhsinst "C_XCL${x}_WRITEXFER"]

    if {[string compare -nocase $subtype "IXCL"] != 0 && [string compare -nocase $subtype "IPLB"] != 0} {

        if {((($basetype == 1) && ($writexfer > 0)) || ($basetype != 1))} { 

            if {[string compare -nocase $fifotype "BRAM"] == 0 || [string compare -nocase $fifotype "SRL"] == 0} {

            return 
            }
        }
    }

    error "\n No writable port - Invalid configuration.\n" "" "mdt_error" 

}


#***--------------------------------***-----------------------------------***
#
#                            CORE_LEVEL_CONSTRAINTS
#
#***--------------------------------***-----------------------------------***


proc generate_corelevel_ucf {mhsinst} {

    set  filePath [xget_ncf_dir $mhsinst]

    file mkdir    $filePath

    # specify file name
    set    instname   [xget_hw_parameter_value $mhsinst "INSTANCE"]
    set    ipname     [xget_hw_option_value    $mhsinst "IPNAME"]
    set    name_lower [string   tolower   $instname]
    set    fileName   $name_lower
    append fileName   "_wrapper.ucf"
    append filePath   $fileName

    # Open a file for writing
    set    outputFile [open $filePath "w"]

    #---------------------------------------------------------------------
    #     BEGIN: V5 DDR2 MIG PHY dependent constraints
    #---------------------------------------------------------------------
    set family      [xget_hw_parameter_value $mhsinst "C_FAMILY"]
    set mem_type    [xget_hw_parameter_value $mhsinst "C_MEM_TYPE"]
    set static_phy  [xget_hw_parameter_value $mhsinst "C_USE_STATIC_PHY"]

    if { [ string compare -nocase $family "virtex5" ] == 0 && [ string compare -nocase $mem_type "DDR2" ] == 0 && $static_phy == 0 } {
        puts $outputFile "AREA_GROUP \"DDR_CAPTURE_FFS\" GROUP = CLOSED;"
    }

    #---------------------------------------------------------------------
    #     BEGIN: C_IDELAYCTRL_LOC dependent constraints
    #---------------------------------------------------------------------

    # The following constraints are parameter C_IDELAYCTRL_LOC dependent
    set loc_values [xget_hw_parameter_value $mhsinst "C_IDELAYCTRL_LOC" ]  
    # the value must be set
    if { [string length $loc_values] == 0 } {

        puts  "\nWARNING:  $instname ($ipname) -\n      The parameter C_IDELAYCTRL_LOC must be set.\n"

    } else {

        # if this is not a default - "NOT_SET", then generate ucf file
        if { [string compare -nocase $loc_values "NOT_SET"] != 0 } {

            # get a list by splitting $loc_values at each "-"
            set vlist       [split $loc_values -]
            set vlist_upper [string toupper $vlist]
            set list_length [llength $vlist_upper]
            set idelay_num  [xget_hw_parameter_value $mhsinst "C_NUM_IDELAYCTRL" ] 

            # The list length must be equal to C_NUM_IDELAYCTRL
            if { [string compare $list_length $idelay_num] == 0 } {

                # check if any invalid pair value
                foreach value $vlist_upper {

                    if { [regexp IDELAYCTRL_X\[0-9\]\?\[0-9\]Y\[0-9\]\?\[0-9\] $value] == 0 } {

                        error "\n Invalid parameter C_IDELAYCTRL_LOC = $loc_values" "" "mdt_error"

                    }

                }

                set suffix 0
                         
                foreach value $vlist_upper {

                    puts   $outputFile "INST \"${instname}/gen_instantiate_idelayctrls\[${suffix}\].idelayctrl0\"   LOC = \"${value}\";"
                    incr suffix

                }  
            } else {

                error "\n parameter C_IDELAYCTRL_LOC must have a list of value equals to C_NUM_IDELAYCTRL" "" "mdt_error"

            }

        }

    }

        # VFBC Specific Constraints

    set portno [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]    

    set vfbc_clock_period_ps [xget_hw_parameter_value $mhsinst C_MPMC_CLK0_PERIOD_PS]

    if {$vfbc_clock_period_ps == ""} {
      set vfbc_clock_port [xget_hw_port_handle $mhsinst "MPMC_Clk0"]
      set vfbc_clock_frequency [xget_hw_subproperty_value $vfbc_clock_port "CLK_FREQ_HZ"]
      
      if {$vfbc_clock_frequency == ""} {
        set vfbc_clock_frequency 133333333
        
#        set C_FAMILY [xget_hw_parameter_value $mhsinst "C_FAMILY"]
        if {[ string compare -nocase $family "virtex4" ] == 0 || [ string compare -nocase $family "qvirtex4" ] == 0 || [ string compare -nocase $family "qrvirtex4" ] == 0} {
          set vfbc_clock_frequency 166666666
        } elseif {[ string compare -nocase $family "virtex5" ] == 0} {
          set vfbc_clock_frequency 200000000
        }
      } 
      set vfbc_clock_period_ps [expr 1000000000000/${vfbc_clock_frequency}]   
    }
    #puts "MPMC (${instname}): Setting VFBC PIM Clock Period to $vfbc_clock_period_ps ps"     

    for {set x 0} {$x < $portno} {incr x 1} {
        set i_btype [xget_hw_parameter_value $mhsinst "C_PIM${x}_BASETYPE"]
        set s_btype [lindex {INACTIVE XCL PLB SDMA MPMC_PIM PPC440MC VFBC} $i_btype]

        if { [string compare -nocase $s_btype "VFBC"] == 0 } {  ## Port is VFBC
          puts $outputFile "Net \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/cmd_reset*\" TIG;"
          puts $outputFile "Net \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/wrfifo_reset*\" TIG;"
          puts $outputFile "Net \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/wrfifo_flush*\" TIG;"
          puts $outputFile "Net \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/rdfifo_reset*\" TIG;"
          puts $outputFile "Net \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/rdfifo_flush*\" TIG;"
          puts $outputFile "Net \"*${instname}/VFBC${x}_INST.vfbc/*/npi_init_done*\" TIG;"

          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/ObjFifo_cons_ctl_comp/ObjPtr*\" TNM = \"${instname}_VFBC${x}_obj_ptr_src_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/ObjFifo_prod_ctl_comp/ObjPtr*\" TNM = \"${instname}_VFBC${x}_obj_ptr_src_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/*ConsObjPtr_sync2prod*\" TNM = \"${instname}_VFBC${x}_obj_ptr_dest_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/*ProdObjPtr_sync2cons*\" TNM = \"${instname}_VFBC${x}_obj_ptr_dest_grp\";"
          puts $outputFile "TIMESPEC \"TS_${instname}_VFBC${x}_Obj_ptr_grp\" = FROM \"${instname}_VFBC${x}_obj_ptr_src_grp\" to \"${instname}_VFBC${x}_obj_ptr_dest_grp\" TIG;"
          
          
          
               
          ### source
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/GEN_BURST_CUT.GEN_MULTI_PORT.UCmdFetch/bc_valid*\" TNM = \"${instname}_VFBC${x}_counter_src_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/GEN_BURST_CUT.UBurstCtrl/x_offset_stage1*\" TNM = \"${instname}_VFBC${x}_counter_src_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/GEN_BURST_CUT.UBurstCtrl/address_out*\" TNM = \"${instname}_VFBC${x}_counter_src_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/GEN_BURST_CUT.UBurstCtrl/o_carry*\" TNM = \"${instname}_VFBC${x}_counter_src_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/GEN_BURST_CUT.UBurstCtrl/x_size_out*\" TNM = \"${instname}_VFBC${x}_counter_src_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/GEN_BURST_CUT.GEN_MULTI_PORT.UCmdBuf/cmd_mem*\" TNM = \"${instname}_VFBC${x}_counter_src_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC_BACKEND_CTRL/Address_fifo0/read_ptr*\" TNM = \"${instname}_VFBC${x}_counter_src_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC_BACKEND_CTRL/Address_fifo0/write_ptr*\" TNM = \"${instname}_VFBC${x}_counter_src_grp\";"
              
          ### Destination
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/GEN_BURST_CUT.UBurstCtrl/x_carry*\" TNM = \"${instname}_VFBC${x}_counter_dest_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/GEN_BURST_CUT.UBurstCtrl/command_done*\" TNM = \"${instname}_VFBC${x}_counter_dest_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC/GEN_BURST_CUT.UBurstCtrl/x_size_lower_next*\" TNM = \"${instname}_VFBC${x}_counter_dest_grp\";"
          puts $outputFile "INST \"*${instname}/VFBC${x}_INST.vfbc/*/UVFBC_BACKEND_CTRL/Address_fifo0/mem1/*\" TNM = \"${instname}_VFBC${x}_counter_dest_grp\";"
          
          puts $outputFile "TIMESPEC \"TS_${instname}_VFBC${x}_counters\" = FROM \"${instname}_VFBC${x}_counter_src_grp\" TO \"${instname}_VFBC${x}_counter_dest_grp\" ${vfbc_clock_period_ps} ps PRIORITY 1;\n"
        }
    }


    # Close the file
    close $outputFile
    puts  [xget_ncf_loc_info $mhsinst]

}


#***--------------------------------***------------------------------------***
#
#                        Update Control Path Parameters 
#
#***--------------------------------***------------------------------------***

#-------------------------------------------
# initial control database for lookup
#-------------------------------------------
proc init_control {mhsinst} {

    global array_ctrl_param
    global array_mpmcinst
    
    set instname [xget_hw_parameter_value $mhsinst "INSTANCE"]
    set mpmclist [array get array_mpmcinst $instname]

    if {[string length $mpmclist] == 0} {

        set array_mpmcinst($instname) $instname
        array unset array_ctrl_param
    } else {

        return
    }

    if { [array size array_ctrl_param] == 0 } {

        check_partno $mhsinst

        # PERL script is called to generate a mpmc_ctrl_path_params.v file
        set    perlPath [xget_hw_pcore_dir $mhsinst]
        append perlPath "generate_ctrl_path_table.pl"

        if { [file exists $perlPath] == 1 && [file readable $perlPath] == 1} {

            if {[file exists __xps] == 0} {
                file mkdir   __xps 
            }

            set    verFile [pwd]
            append verFile "/__xps/${instname}_mpmc_ctrl_path_params.v"
            set    txtFile [pwd]
            append txtFile "/__xps/${instname}_mpmc_ctrl_path_table.txt"
            set    errFile [pwd]
            append errFile "/__xps/${instname}_ctrl_path_generation_errors.txt"

            # execute perl script
        exec xilperl $perlPath \
            -mpmc3                                                                              \
            -family      [xget_hw_parameter_value $mhsinst "C_FAMILY"]                          \
            -enable_ecc  [xget_hw_parameter_value $mhsinst "C_INCLUDE_ECC_SUPPORT"]             \
            -tRAS        [expr [xget_hw_parameter_value $mhsinst "C_MEM_PART_TRAS"]/1000.0]     \
            -tRASmax     [expr [xget_hw_parameter_value $mhsinst "C_MEM_PART_TRASMAX"]/1000.0]  \
            -part_data_width [xget_hw_parameter_value $mhsinst "C_MEM_PART_DATA_WIDTH"]         \
            -CAS_Latency_A_Fmax [xget_hw_parameter_value $mhsinst "C_MEM_PART_CAS_A_FMAX"]      \
            -CAS_Latency_A      [xget_hw_parameter_value $mhsinst "C_MEM_PART_CAS_A"]           \
            -CAS_Latency_B_Fmax [xget_hw_parameter_value $mhsinst "C_MEM_PART_CAS_B_FMAX"]      \
            -CAS_Latency_B      [xget_hw_parameter_value $mhsinst "C_MEM_PART_CAS_B"]           \
            -CAS_Latency_C_Fmax [xget_hw_parameter_value $mhsinst "C_MEM_PART_CAS_C_FMAX"]      \
            -CAS_Latency_C      [xget_hw_parameter_value $mhsinst "C_MEM_PART_CAS_C"]           \
            -CAS_Latency_D_Fmax [xget_hw_parameter_value $mhsinst "C_MEM_PART_CAS_D_FMAX"]      \
            -CAS_Latency_D      [xget_hw_parameter_value $mhsinst "C_MEM_PART_CAS_D"]           \
            -memory_burst_length 4                                                              \
            -memory_num_ranks   [xget_hw_parameter_value $mhsinst "C_MEM_NUM_RANKS"]            \
            -tRC    [expr [xget_hw_parameter_value $mhsinst "C_MEM_PART_TRC"]/1000.0]           \
            -tRCD   [expr [xget_hw_parameter_value $mhsinst "C_MEM_PART_TRCD"]/1000.0]          \
            -tDQSS  [xget_hw_parameter_value $mhsinst "C_MEM_PART_TDQSS"]                       \
            -tWR    [expr [xget_hw_parameter_value $mhsinst "C_MEM_PART_TWR"]/1000.0]           \
            -tRP    [expr [xget_hw_parameter_value $mhsinst "C_MEM_PART_TRP"]/1000.0]           \
            -tMRD   [xget_hw_parameter_value $mhsinst "C_MEM_PART_TMRD"]                        \
            -tRRD   [expr [xget_hw_parameter_value $mhsinst "C_MEM_PART_TRRD"]/1000.0]          \
            -tRFC   [expr [xget_hw_parameter_value $mhsinst "C_MEM_PART_TRFC"]/1000.0]          \
            -tAL    [xget_hw_parameter_value $mhsinst "C_MEM_PART_TAL"]                         \
            -tCCD   [xget_hw_parameter_value $mhsinst "C_MEM_PART_TCCD"]                        \
            -tWTR   [expr [xget_hw_parameter_value $mhsinst "C_MEM_PART_TWTR"]/1000.0]          \
            -tRTP   [expr [xget_hw_parameter_value $mhsinst "C_MEM_PART_TRTP"]/1000.0]          \
            -c      [expr 1000000.0/[xget_hw_parameter_value $mhsinst C_MPMC_CLK0_PERIOD_PS]]   \
            -reg    [xget_hw_parameter_value $mhsinst "C_MEM_REG_DIMM"]                         \
            -m      [xget_hw_parameter_value $mhsinst "C_MEM_TYPE"]                             \
            -d      [xget_hw_parameter_value $mhsinst "C_MEM_DATA_WIDTH"]                       \
            -f_txt  $txtFile                                                                    \
            -f_err  $errFile                                                                    \
            -f_ver  $verFile 

            if { [file exists $verFile] == 1 && [file readable $verFile] == 1} {

                parsePerlFile $verFile

            } else {

                error "\n File $verFile is not found or unable to open for reading.\n" "" "mdt_error" 

            }


        } else {

            error "\n File $perlPath is not found or unable to open for reading.\n" "" "mdt_error" 

        }
    }
}

#-------------------------------------------
# - parse the mpmc_ctrl_path_params.v file 
# - create database array_ctrl_param 
#-------------------------------------------
proc parsePerlFile {filePath}  {

    global array_ctrl_param

    set infID       [open $filePath r]
    
    while {[gets $infID line] >= 0 } {

        if {[string length $line] == 0 || [string first "//" $line 0] != -1} {

            continue
        }

        if {[string match -nocase PARAMETER* $line] == 1} {

            set line_trim    [string trim $line ";"]
            set list [split  $line_trim   " "] 
            set param_name   ""
            set param_value  ""
            set size         [llength $list]

            # create database for array_partno_param
            for {set i 0} {$i < $size} {incr i 1} {
            
                set item [lindex $list $i]

                if { [string match -nocase C_* $item] == 1 } {
                    
                    set param_name $item
                    break
                } 

            }

            # get the parameter value
            set vlog_value [lindex $list [expr $size - 1]]

            if {[string match "*'h*" $vlog_value] == 1} {

                set hex_value [lindex [split $vlog_value "'h"] 2]
                set array_ctrl_param($param_name) [format "0x%s" $hex_value]

            } elseif {[string match "*'d*" $vlog_value] == 1} {

                set dec_value [lindex [split $vlog_value "'d"] 2]
                set array_ctrl_param($param_name) $dec_value

            } else {

                set array_ctrl_param($param_name) $vlog_value

            }
        }
    }

    close $infID

}


#-------------------------------------------
# - Compute the HIGHADDR based on BASEADDR 
# - Compute C_NUM_PORTS 
#-------------------------------------------
proc xps_ipconfig_accept {mhsinst} {

    # C_NUM_PORTS
    update_portno        $mhsinst

    # C_SDMA_CTRL_HIGHADDR
    update_sdma_ctrl_highaddr $mhsinst

    # C_MPMC_CTRL_HIGHADDR
    # The update of MPMC Control High address will be done when the base address is changed
    # The UI file will directly make the call

    # update_mpmc_ctrl_highaddr $mhsinst

}

# C_NUM_PORTS
# find the highest active port + 1
proc update_portno {mhsinst} {

    for {set x 7} {$x >= 0} {incr x -1} {

        set basetype [xget_hw_parameter_value $mhsinst "C_PIM${x}_BASETYPE"]

        if {$basetype != 0} {

            set parm_handle [xget_hw_parameter_handle $mhsinst "C_NUM_PORTS"]
            xset_hw_parameter_value $parm_handle [expr $x + 1]
            return
        }
    }

    xset_hw_parameter_value $parm_handle 0
}

proc compute_mem_size {mhsinst} {
    set dwidth   [xget_hw_parameter_value $mhsinst "C_MEM_DATA_WIDTH"]
    set bankbits [xget_hw_parameter_value $mhsinst "C_MEM_PART_NUM_BANK_BITS"]
    set rowbits  [xget_hw_parameter_value $mhsinst "C_MEM_PART_NUM_ROW_BITS"]
    set colbits  [xget_hw_parameter_value $mhsinst "C_MEM_PART_NUM_COL_BITS"]
    set dimmsno  [xget_hw_parameter_value $mhsinst "C_MEM_NUM_DIMMS"]
    set ranksno  [xget_hw_parameter_value $mhsinst "C_MEM_NUM_RANKS"]
    set exponent [expr $bankbits + $rowbits + $colbits + $dimmsno + $ranksno - 2]

    if {[string length $dwidth] == 0 || [string length $bankbits] == 0 
            || [string length $rowbits] == 0 || [string length $colbits] == 0 
            || [string length $dimmsno] == 0 || [string length $ranksno] == 0} {
        return 0
    }

    # this will give you the memory size in Bytes
    set m_size   [format "0x%08x" [expr wide($dwidth/8*(pow(2,$exponent)))]]

    return $m_size
}

proc addresses_not_set {mhsinst} {
    set base [xget_hw_parameter_value $mhsinst "C_MPMC_BASEADDR"]
    set high [xget_hw_parameter_value $mhsinst "C_MPMC_HIGHADDR"]

    if { [expr {$high == 0x0}] && [expr {$base == 0xffffffff}] } {
            return 1
    }

    return 0
}

proc is_aligned { base size } {

        if { [expr $base & [expr $size - 1]] } {
                return 0;
        }
        return 1;
}

proc update_mpmc_highaddr {mhsinst} {
    # compute memory size from parameters in MHS
    set m_size   [compute_mem_size $mhsinst]

    # print out size in MB units in the GUI
    set msize_handle [xget_hw_parameter_handle $mhsinst "G_MEM_SIZE"]
    set convertor    [expr int(pow(2,20))]
    set final_val    [expr $m_size/$convertor]
    append           final_val "MB"
    xset_hw_parameter_value $msize_handle $final_val

    # if MPMC has been added for the first time, addresses will not be set
    # in this case we assign base = 0
    set baseaddr 0
    if { [addresses_not_set $mhsinst] } {
        set baseaddr 0
    } else {
        set baseaddr_str [xget_hw_parameter_value $mhsinst "C_MPMC_BASEADDR"]
        set baseaddr     [xformat_address_string  $baseaddr_str]
    }

    # check to see if we have alignment/overflow issues
    set overflow_error 0
    set aligned_error 0
    
    ## The and with 0xFFFFFFFF is required in order to ensure that we
    ## support 64 bit machines. The addresses dont wrap around on that.

    if { [expr $baseaddr >  [expr [expr $baseaddr + $m_size - 1] & 0xffffffff ]] } {
        set baseaddr 0
        set overflow_error 1
    } elseif { ![is_aligned $baseaddr $m_size] } {
        set baseaddr 0
        set aligned_error 1
    }

    set highaddr [expr $baseaddr + $m_size - 1]

    # set C_MPMC_HIGHADDR
    set high_handle [xget_hw_parameter_handle $mhsinst "C_MPMC_HIGHADDR"]
    xset_hw_parameter_value $high_handle [format "0x%08X" $highaddr]

    # set C_MPMC_BASEADDR
    set base_handle [xget_hw_parameter_handle $mhsinst "C_MPMC_BASEADDR"]
    xset_hw_parameter_value $base_handle [format "0x%08X" $baseaddr]

    if { $overflow_error } {
        error "\nAuto-computed high address overflows the 32-bit address space. Setting base address to 0x00000000.\n" "" "mdt_error"
    }

    if { $aligned_error } {
        #error "\nBase address not aligned on memory size boundary. Setting base address to 0x00000000.\n" "" "mdt_error"
        error "\nParameter changes affected the Base and High address of this memory system. The base address is not aligned to the memory size in an even power of 2. The Base Address has been reset to 0x00000000.\n" "" "mdt_error"
    }
}


# C_SDMA_CTRL_HIGHADDR = 64k (0x10000) + C_SDMA_CTRL_BASEADDR
# future enhancement - Add a check for overflow & alignment 

proc update_sdma_ctrl_highaddr {mhsinst} {

    set baseaddr    [xget_hw_parameter_value  $mhsinst "C_SDMA_CTRL_BASEADDR"]

    # format address string to hex
    set baseaddr    [xformat_address_string  $baseaddr]
    set highaddr    [expr $baseaddr + 0x10000 - 1]
    set highaddr    [xconvert_num_to_binary $highaddr 8]
    set highaddr    [xconvert_binary_to_hex $highaddr]

    set high_handle [xget_hw_parameter_handle $mhsinst "C_SDMA_CTRL_HIGHADDR"]
    xset_hw_parameter_value $high_handle $highaddr
}

# C_MPMC_CTRL_HIGHADDR = 64k (0x10000) + C_MPMC_CTRL_BASEADDR
proc update_mpmc_ctrl_highaddr {mhsinst} {

    set baseaddr    [xget_hw_parameter_value  $mhsinst "C_MPMC_CTRL_BASEADDR"]

    # format address string to hex
    set baseaddr    [xformat_address_string  $baseaddr]
    set highaddr    [expr $baseaddr + 0x10000 - 1]
    set highaddr    [xconvert_num_to_binary $highaddr 8]
    set highaddr    [xconvert_binary_to_hex $highaddr]

    set high_handle [xget_hw_parameter_handle $mhsinst "C_MPMC_CTRL_HIGHADDR"]
    xset_hw_parameter_value $high_handle $highaddr
}

#-------------------------------------------
# if C_ALL_PIMS_SHARE_ADDRESSES = 0, then
#   - set C_PIMx_BASEADDR with C_MPMC_BASEADDR 
#   - set C_SDMA_CTRLx_BASEADDR with C_SDMA_CTRL_BASEADDR 
#-------------------------------------------
proc mpmc_copy_port_addr {mhsinst} {

    set share_addr [xget_hw_parameter_value $mhsinst "C_ALL_PIMS_SHARE_ADDRESSES"]

    if {$share_addr != 0 } {

        return 
    }

    set mpmc_baseaddr [xget_hw_parameter_value $mhsinst "C_MPMC_BASEADDR"]
    set mpmc_highaddr [xget_hw_parameter_value $mhsinst "C_MPMC_HIGHADDR"]
    
    set sdma_baseaddr [xget_hw_parameter_value $mhsinst "C_SDMA_CTRL_BASEADDR"]
    set sdma_highaddr [xget_hw_parameter_value $mhsinst "C_SDMA_CTRL_HIGHADDR"]

    for {set x 0} {$x < 8} {incr x 1} {
        
        set pim_basehandle  [xget_hw_parameter_handle $mhsinst "C_PIM${x}_BASEADDR"]
        set pim_highhandle  [xget_hw_parameter_handle $mhsinst "C_PIM${x}_HIGHADDR"]

        set sdma_basehandle [xget_hw_parameter_handle $mhsinst "C_SDMA_CTRL${x}_BASEADDR"]
        set sdma_highhandle [xget_hw_parameter_handle $mhsinst "C_SDMA_CTRL${x}_HIGHADDR"]
        xset_hw_parameter_value $pim_basehandle  $mpmc_baseaddr
        xset_hw_parameter_value $pim_highhandle  $mpmc_highaddr
        xset_hw_parameter_value $sdma_basehandle $sdma_baseaddr
        xset_hw_parameter_value $sdma_highhandle $sdma_highaddr

    }
}


#-------------------------------------------
# Left Justify
# Scan through all C_PIM*_BASETYPE parameters, 
# if there is any hole between active ports,
# then shift ports left to eliminate the holes.
#-------------------------------------------
proc mpmc_left_justify {mhsinst} {

    set last 0

    for {set index 0} {$index < 8 } {incr index 1} {

        set index_handle [xget_hw_parameter_handle $mhsinst C_PIM${index}_BASETYPE]
        set index_value  [xget_hw_value            $index_handle]

        if {$index_value != 0} {

            if {$last != $index} {

                set last_handle  [xget_hw_parameter_handle $mhsinst C_PIM${last}_BASETYPE]
                set last_value   [xget_hw_value            $last_handle]

                # swap C_PIM${index}_BASETYPE with C_PIM${last}_BASETYPE
                xset_hw_parameter_value  $index_handle $last_value
                xset_hw_parameter_value  $last_handle  $index_value

                # shift parameters
                shift_parameters $mhsinst $index $last 

                # shift external connectors
                shift_external_connectors $mhsinst $index $last 

            }
            incr last 1

        }
    }
}

# shift port specific paramters
proc shift_parameters {mhsinst index last} {

    swap_parameter_values $mhsinst "C_XCL"  "_WRITEXFER"       $index $last
    swap_parameter_values $mhsinst "C_XCL"  "_LINESIZE"         $index $last         
    swap_parameter_values $mhsinst "C_SPLB" "_NATIVE_DWIDTH"    $index $last   
    swap_parameter_values $mhsinst "C_SDMA" "_COMPLETED_ERR_TX" $index $last
    swap_parameter_values $mhsinst "C_SDMA" "_COMPLETED_ERR_RX" $index $last
    swap_parameter_values $mhsinst "C_SDMA" "_PRESCALAR"        $index $last        
    swap_parameter_values $mhsinst "C_SDMA" "_PI2LL_CLK_RATIO"  $index $last  
    # 
    swap_parameter_values $mhsinst "C_PIM"       "_BASEADDR"    $index $last    
    swap_parameter_values $mhsinst "C_PIM"       "_HIGHADDR"    $index $last    
    swap_parameter_values $mhsinst "C_PIM"       "_OFFSET"      $index $last      
    swap_parameter_values $mhsinst "C_SDMA_CTRL" "_BASEADDR"    $index $last    
    swap_parameter_values $mhsinst "C_SDMA_CTRL" "_HIGHADDR"    $index $last    
    #
    swap_parameter_values $mhsinst "C_PPC440MC" "_BURST_LENGTH"    $index $last
    swap_parameter_values $mhsinst "C_PPC440MC" "_PIPE_STAGES"     $index $last
    #
    #swap_parameter_values $mhsinst "C_VFBC" "_BURST_LENGTH"        $index $last
    #swap_parameter_values $mhsinst "C_VFBC" "_ASYNC_CLOCK"         $index $last
    swap_parameter_values $mhsinst "C_VFBC" "_CMD_FIFO_DEPTH"      $index $last
    swap_parameter_values $mhsinst "C_VFBC" "_CMD_AFULL_COUNT"       $index $last
    #swap_parameter_values $mhsinst "C_VFBC" "_WD_ENABLE"           $index $last
    #swap_parameter_values $mhsinst "C_VFBC" "_WD_DATA_WIDTH"       $index $last
    #swap_parameter_values $mhsinst "C_VFBC" "_WD_FIFO_DEPTH"       $index $last
    #swap_parameter_values $mhsinst "C_VFBC" "_WD_AFULL_COUNT"      $index $last
    #swap_parameter_values $mhsinst "C_VFBC" "_RD_ENABLE"           $index $last
    swap_parameter_values $mhsinst "C_VFBC" "_RDWD_DATA_WIDTH"       $index $last
    swap_parameter_values $mhsinst "C_VFBC" "_RDWD_FIFO_DEPTH"       $index $last
    swap_parameter_values $mhsinst "C_VFBC" "_RD_AEMPTY_WD_AFULL_COUNT"     $index $last
    #
    swap_parameter_values $mhsinst "C_PIM" "_DATA_WIDTH"           $index $last         
    swap_parameter_values $mhsinst "C_PI"  "_RD_FIFO_TYPE"         $index $last         
    swap_parameter_values $mhsinst "C_PI"  "_WR_FIFO_TYPE"         $index $last         
    swap_parameter_values $mhsinst "C_PI"  "_RD_FIFO_MEM_PIPELINE" $index $last
    swap_parameter_values $mhsinst "C_PI"  "_RD_FIFO_APP_PIPELINE" $index $last
    swap_parameter_values $mhsinst "C_PI"  "_WR_FIFO_MEM_PIPELINE" $index $last
    swap_parameter_values $mhsinst "C_PI"  "_WR_FIFO_APP_PIPELINE" $index $last
    swap_parameter_values $mhsinst "C_PI"  "_ADDRACK_PIPELINE"     $index $last
    #
    swap_parameter_values $mhsinst "C_PI"  "_PM_USED"     $index $last
    swap_parameter_values $mhsinst "C_PI"  "_PM_DC_CNTR"  $index $last

}

# shift port specific external connectors
# XCLx, SPLBx, SDMA_CTRLx, SDMA_LLx, MPMC_PIMx, PPC440MCx, VFBCx
proc shift_external_connectors {mhsinst index last} {

    swap_connectors $mhsinst "XCL"       $index $last
    swap_connectors $mhsinst "SPLB"      $index $last
    swap_connectors $mhsinst "SDMA_CTRL" $index $last
    swap_connectors $mhsinst "SDMA_LL"   $index $last
    swap_connectors $mhsinst "MPMC_PIM"  $index $last
    swap_connectors $mhsinst "PPC440MC"  $index $last
    swap_connectors $mhsinst "VFBC"      $index $last


}


proc swap_connectors {mhsinst bus index last} {

    set last_bus_connector  [xget_hw_busif_value $mhsinst $bus$last]
    set index_bus_connector [xget_hw_busif_value $mhsinst $bus$index]

    xadd_hw_ipinst_busif    $mhsinst $bus$index  "No Connection"
    xadd_hw_ipinst_busif    $mhsinst $bus$last   $index_bus_connector

}

proc swap_parameter_values {mhsinst head tail index last} {

    set last_name    $head$last$tail
    set last_handle  [xget_hw_parameter_handle $mhsinst $last_name]
    set last_value   [xget_hw_value            $last_handle]

    set index_name   $head$index$tail
    set index_handle [xget_hw_parameter_handle $mhsinst $index_name]
    set index_value  [xget_hw_value            $index_handle]

    xset_hw_parameter_value  $index_handle $last_value
    xset_hw_parameter_value  $last_handle  $index_value

}


#-------------------------------------------
# Restore following parameters to their default value 
#-------------------------------------------

proc mpmc_restore_default_addr {mhsinst} {

    restore_param_default [xget_hw_parameter_handle  $mhsinst "C_MPMC_BASEADDR"]
    restore_param_default [xget_hw_parameter_handle  $mhsinst "C_SDMA_CTRL_BASEADDR"]
    restore_param_default [xget_hw_parameter_handle  $mhsinst "C_ALL_PIMS_SHARE_ADDRESSES"]

}

proc mpmc_restore_default_datapath {mhsinst} {

    for {set x 0} {$x < 8 } {incr x 1} {
 
        restore_param_default [xget_hw_parameter_handle  $mhsinst "C_PIM${x}_DATA_WIDTH"]
    }

    for {set x 0} {$x < 8 } {incr x 1} {
 
        restore_param_default [xget_hw_parameter_handle  $mhsinst "C_PI${x}_RD_FIFO_TYPE"]
    }

    for {set x 0} {$x < 8 } {incr x 1} {
 
        restore_param_default [xget_hw_parameter_handle  $mhsinst "C_PI${x}_WR_FIFO_TYPE"]
    }

    for {set x 0} {$x < 8 } {incr x 1} {
 
        restore_param_default [xget_hw_parameter_handle  $mhsinst "C_PI${x}_RD_FIFO_MEM_PIPELINE"]
    }

    for {set x 0} {$x < 8 } {incr x 1} {
 
        restore_param_default [xget_hw_parameter_handle  $mhsinst "C_PI${x}_RD_FIFO_APP_PIPELINE"]
    }

    for {set x 0} {$x < 8 } {incr x 1} {
 
        restore_param_default [xget_hw_parameter_handle  $mhsinst "C_PI${x}_WR_FIFO_MEM_PIPELINE"]
    }

    for {set x 0} {$x < 8 } {incr x 1} {
 
        restore_param_default [xget_hw_parameter_handle  $mhsinst "C_PI${x}_WR_FIFO_APP_PIPELINE"]
    }

    for {set x 0} {$x < 8 } {incr x 1} {
 
        restore_param_default [xget_hw_parameter_handle  $mhsinst "C_PI${x}_ADDRACK_PIPELINE"]
    }

}

proc mpmc_restore_default_arb {mhsinst} {

    restore_param_default [xget_hw_parameter_handle  $mhsinst "C_ARB0_ALGO"]
    restore_param_default [xget_hw_parameter_handle  $mhsinst "C_ARB_PIPELINE"]
    mpmc_update_arbitration  $mhsinst

}


proc mpmc_restore_default_debug {mhsinst} {

    restore_param_default [xget_hw_parameter_handle  $mhsinst "C_PM_ENABLE"]
    restore_param_default [xget_hw_parameter_handle  $mhsinst "C_INCLUDE_ECC_SUPPORT"]
    restore_param_default [xget_hw_parameter_handle  $mhsinst "C_DEBUG_REG_ENABLE"]

}

proc restore_param_default {param_handle} {

    set param_default [xget_hw_subproperty_value $param_handle "MPD_VALUE"]
    xset_hw_parameter_value  $param_handle $param_default

}


####
#
# load_memory_parameters
#
# Loads all memory parameters from the database. It does call INIT to
# check if the database is initialized or not
#
# Called by : XPS GUI Only. No backend (platgen or DRC call)
# 
###

proc load_memory_parameters {mhsinst forceOverwrite} {

    global array_partno_param
    global array_partnolist

    init_memory $mhsinst
    
    set partno [string toupper [xget_hw_parameter_value  $mhsinst "C_MEM_PARTNO"]]

    
    if { [string compare -nocase $partno "CUSTOM"] == 0 ||  [string compare -nocase $partno "NONE"] == 0  } {
        return
    } 

    set list   [array get array_partno_param $partno,*]

    foreach {key value}  $list {

        set param  [lindex [split $key ,] 1]

        if { [string match -nocase C_MEM* $param] == 1 } {

            set param_handle [xget_hw_parameter_handle  $mhsinst $param]
            set mhs_value    [xget_hw_subproperty_value $param_handle "MHS_VALUE"]

            if { $forceOverwrite == 1 || [string length $mhs_value] == 0 } {
            xset_hw_parameter_value  $param_handle $value
            }
        }
    }


    # update VALUES tag to G_MEM_TYPE
    set list [array get array_partnolist $partno,*,*,*,*,*]
    set key  [lindex $list 0]
    set type [lindex [split $key ,] 2]

    set param_handle [xget_hw_parameter_handle $mhsinst "C_MEM_TYPE"]
    xset_hw_parameter_value $param_handle $type

}


#-------------------------------------------
# given Style, Type, Density, Part_Width, and Manufactuer
# update all memory path parameters
#-------------------------------------------
proc filter_partno {mhsinst} {

    global array_partnolist

    init_memory $mhsinst

    set Style       [xget_hw_parameter_value $mhsinst "G_MEM_STYLE"]
    set Type        [xget_hw_parameter_value $mhsinst "G_MEM_TYPE"]
    set Density     [xget_hw_parameter_value $mhsinst "G_MEM_DENSITY"]
    set Part_Width  [xget_hw_parameter_value $mhsinst "G_MEM_WIDTH"]
    set Manufactuer [xget_hw_parameter_value $mhsinst "G_MEM_MANUFACTURE"]
    set list        "" 
    
    if {[string length $Style] == 0 || [string length $Type] == 0 || [string length $Density] == 0 || [string length $Part_Width] == 0 || [string length $Manufactuer] == 0} { 

        error "\n The information for Style, Type, Density, Part_Width, or Manufactuer is incomplete.\n" "" "mdt_error" 

    } 

    # update VALUES tag to C_MEM_PARTNO
    set list [array get array_partnolist *,$Style,$Type,$Density,$Part_Width,$Manufactuer]

    set list_unsorted ""

    foreach {key value}  $list {

        lappend list_unsorted $value 

    }

    set list_sorted [lsort $list_unsorted]
    set list_partno "(Select a part= Select a part, CUSTOM= CUSTOM "

    for {set i 0} {$i < [llength $list_sorted]} {incr i 1} {

        append list_partno ", "
        append list_partno "[lindex $list_sorted $i]= [lindex $list_sorted $i]" 

    }

    append list_partno ")" 

    set param_handle    [xget_hw_parameter_handle $mhsinst "G_MEM_PARTNO"]
    xadd_hw_subproperty $param_handle             "VALUES" $list_partno

    # set the value to "Please select a part"
    xset_hw_parameter_value $param_handle "Select a part"
}

#-------------------------------------------
# initial memory database for lookup
#-------------------------------------------
proc init_memory {mhsinst} {

    global array_partno_param
    global array_partnolist

    if { [array size array_partno_param] == 0 || [array size array_partnolist] == 0 } {

        # specify file name
        set    filePath [xget_hw_pcore_dir $mhsinst]
        append filePath "mpmc_memory_database.csv"

        if { [file exists $filePath] == 1 && [file readable $filePath] == 1} {

            parseCSVFile    $filePath

        } else {

            error "\n File $filePath is not found or unable to open for reading.\n" "" "mdt_error" 

        }
    }
}

#
# parse the CSV file
#
#-------------------------------------------
# - parse the mpmc_memory_database.csv file 
# - create database array_partno_param & array_partnolist 
#-------------------------------------------
proc parseCSVFile {filename}  {

    global array_partno_param
    global array_partnolist
    set infID       [open $filename r]
    set listNames   ""
    
    while {[gets $infID line] >= 0 } {

        if { [llength $listNames] == 0 } {

            # split on commas
            set listNames [split $line ,]

        } else {

            set listData [split $line ,]
            set partno   [lindex $listData [lsearch $listNames "C_MEM_PARTNO"]]
            set partno   [string toupper   $partno]
            set style    [lindex $listData [lsearch $listNames "Style"]]
            set type     [lindex $listData [lsearch $listNames "Type"]]
            set density  [lindex $listData [lsearch $listNames "Density"]]
            set pwidth   [lindex $listData [lsearch $listNames "Part_Width"]]
            set manfctr  [lindex $listData [lsearch $listNames "Manufactuer"]]

            # create database for array_partnolist
            set array_partnolist($partno,$style,$type,$density,$pwidth,$manfctr) $partno

            # create database for array_partno_param
            for {set i 0} {$i < [llength $listData]} {incr i 1} {
            
                set param_name [string toupper [lindex $listNames $i]]

                if { [string match -nocase C_MEM* $param_name] == 1} {
                    
                    set array_partno_param($partno,$param_name) [lindex $listData $i]
                }

            }

        }

    }

    close $infID

}


#######################################################################
#
# Functions triggerd by the GUI
#
# update_memory_parameters -->  when customer selects a part in the GUI
# filter_partno --> used when customer select the Mem Type, Density and other non MPD params
# update_mpmc_highaddr --> Change in memory triggers High address Update
# mpmc_left_justify --> When Left Justify button is clicked
# mpmc_copy_portaddr --> Copy shared address to non-shared address
# xps_ipconfig_init --> Called by the GUI, when it is stared for the first time
#
#
# Common Helper function called by both backend TCL and GUI
# 
#  --> init_memory
#
#######################################################################

#-------------------------------------------
# called by MPMC Gui before gui is up
#-------------------------------------------
proc xps_ipconfig_init {mhsinst} {

    gui_init $mhsinst
    update_memory_parameters $mhsinst
    filter_partno            $mhsinst 
    mpmc_update_arbitration  $mhsinst
}

proc gui_init {mhsinst} {
    set family      [xget_hw_parameter_value $mhsinst "C_FAMILY"]
    set subfamily      [xget_hw_parameter_value $mhsinst "C_SUBFAMILY"]

    for {set x 0} {$x < 8} {incr x 1} {
        set param_name "C_PIM${x}_BASETYPE"
        set param_handle    [xget_hw_parameter_handle $mhsinst $param_name]

        if { [ string compare -nocase $family "virtex2p" ] == 0 } {
            xadd_hw_subproperty $param_handle "VALUES" \
                "(0=INACTIVE, 1=XCL, 2=PLBV46, 3=SDMA, 4=NPI, 5=PPC440MC-NA, 6=VFBC-NA)"
        } elseif { [ string compare -nocase $family "virtex5" ] == 0 && \
                   [ string compare -nocase $subfamily "fx" ] == 0 } {
        } else {
            xadd_hw_subproperty $param_handle "VALUES" \
                "(0=INACTIVE, 1=XCL, 2=PLBV46, 3=SDMA, 4=NPI, 5=PPC440MC-NA, 6=VFBC)"
        }
    }
}
#-------------------------------------------
# given C_MEM_PARTNO and                
# update all memory path parameters  
# Only called by XPS GUI
#-------------------------------------------
proc update_memory_parameters {mhsinst} {

    set param_handle [xget_hw_parameter_handle $mhsinst "C_MEM_PARTNO"]
    set g_value      [xget_hw_parameter_value  $mhsinst "G_MEM_PARTNO"]
    set c_value      [xget_hw_parameter_value  $mhsinst "C_MEM_PARTNO"]
    set partno       [string toupper $g_value]

    if { [string compare -nocase $partno "Select a part"] == 0} {
        return
    }

    set forceOverwrite 0
    if { [ string compare -nocase $g_value "CUSTOM"] != 0 &&
         [ string compare -nocase $c_value "CUSTOM"] == 0 } {
    set forceOverwrite 1
    }
    
    #--------------------------------------------
    #
    # The value from G_MEM_PARTNO should not be set in the C_MEM_PARTNO during the init phase
    # The value of C_MEM_TYPE is also set into the filter at init stage
    #
    #--------------------------------------------

    if { [string compare -nocase $g_value "NONE"] == 0 } {
        ## Copy the C_MEM_TYPE to G_MEM_TYPE during INIT Phase
        set c_mem_type_value  [xget_hw_parameter_value $mhsinst "C_MEM_TYPE"]
        set g_mem_type_handle [xget_hw_parameter_handle $mhsinst "G_MEM_TYPE"]
        xset_hw_parameter_value $g_mem_type_handle $c_mem_type_value
    } else {
        ## During INIT, do not copy the G_MEM_PARTNO to C_MEM_PARTNO
        xset_hw_parameter_value $param_handle $partno
    }

    load_memory_parameters $mhsinst $forceOverwrite
    update_mpmc_highaddr $mhsinst
}

proc mpmc_update_arbitration {mhsinst} {
    set algo [xget_hw_parameter_value $mhsinst "C_ARB0_ALGO"]
    set portno [xget_hw_parameter_value $mhsinst "C_NUM_PORTS"]
    set numslots [xget_hw_parameter_handle $mhsinst "C_ARB0_NUM_SLOTS"]

    if { [string compare -nocase $algo "ROUND_ROBIN"] == 0 } {
        xset_hw_parameter_value $numslots $portno
        for {set x 0} {$x<16} {incr x 1} {
            set param_name "C_ARB0_SLOT$x"
            set param_handle [xget_hw_parameter_handle $mhsinst $param_name]
            xset_hw_parameter_value $param_handle [mpmc_generate_arbitration_string $portno $x]
        }
    } elseif { [string compare -nocase $algo "FIXED"] == 0 } {
        xset_hw_parameter_value $numslots 1
        set param_handle [xget_hw_parameter_handle $mhsinst "C_ARB0_SLOT0"]
        xset_hw_parameter_value $param_handle [mpmc_generate_arbitration_string $portno 0]
        for {set x 1} {$x<16} {incr x 1} {
            set param_name "C_ARB0_SLOT$x"
            set param_handle [xget_hw_parameter_handle $mhsinst $param_name]
            xset_hw_parameter_value $param_handle ""
        }
    }
}

proc mpmc_generate_arbitration_string {number start} {
    set listDigit ""
    if { $start < $number } {
        for {set x 0} {$x < $number} {incr x 1} {
            set digit [ expr $x + $start ]
            if { $digit >= $number } {
                set digit [ expr $digit - $number ]
            } 
                append listDigit $digit
        }
    }
    return $listDigit
}
######################################################################
#
# Arrays required for storing values & data
#
######################################################################




array set array_mpmcinst  ""

#-----------------------------------
# array_ctrl_param:   
# (key  value) = ($param_name   $param_value)
#-----------------------------------
array set array_ctrl_param   ""

#-----------------------------------
# array_partno_param: 
# (key  value) = ($partno_value,$param_value $param_value)
#-----------------------------------
array set array_partno_param ""

#-----------------------------------
# array_partnolist:   
# (key  value) = ($partno,$Style,$Type,$Density,
#                 $Part_Width,$Manufactuer       $partno )
#-----------------------------------
array set array_partnolist   ""

