#! /usr/local/gnu/bin/perl5 -w
use POSIX qw(ceil floor);
use strict;
use Symbol;
#use filehandle;
#-----------------------------------------------------------------------------
# Copyright (c) 2007 Xilinx, Inc.
# This design is confidential and proprietary of Xilinx, All Rights Reserved.
#-----------------------------------------------------------------------------
#   ____  ____
#  /   /\/   /
# /___/  \  /   Vendor: Xilinx
# \   \   \/    Version: 
#  \   \         
#  /   /        Date Last Modified: Nov 7 2007
# /___/   /\    Date Created:  Aug 10 2005
# \   \  /  \
#  \___\/\___\
# 
#-----------------------------------------------------------------------------
# Purpose: This program generates a text file that describes the MPMC Control 
#          Path BRAM contents.
#
#-----------------------------------------------------------------------------
# Author: JL
# History: 
#   08/10/2005: Initial Version
#   01/17/2006: Complete re-write to clean up and allow for DDR2 corrections
#   01/25/2006: Needed to add a plus 1 to the DDR2 Ctrl_DP_RdFIFO_Push delay. 
#   02/07/2006: Needed to add a plus 1 to C_CTRL_DP_RDFIFO_WHICHPORT_DELAY.
#   07/18/2006: Added command line parameters tRASmax, tRRD, tCCD and tWTR.  
#               tWTR is not used since we don't do any write to reads. 
#               Use strict has been enabled (requiring us to also use Symbol 
#               for filehandles.)  --CC
#   09/05/2006: The plus 1 for DDR2 Ctrl_DP_RdFIFO_Push delay and 
#               C_CTRL_DP_RDFIFO_WHICHPORT_DELAY in DDR2 designs and lack of 
#               the plus 1 in DDR designs will not always be correct.  
#               Frequency, board trace length, and other parameters may affect
#               this.  See release notes for details.
#   09/05/2006: Added extra bit to control path bram state machine to support
#               DDR2 ODT.
#   12/12/2006: Made modifications to use new V5 MIG Phy IF.
#               Removed initialization sequence.
#               We will need to review the DQS_O and rdfifo_push delays if we
#               decide to remove the address path pipeline stages.
#   02/23/2007: Made modifications to correct delays for new MIG phy.
#   03/12/2007: Added ECC support.
#   05/07/2007: Added future support for mpmc3
#
#--------------------------------------------------------------------------
#
# General Rules:
#   1. Read sequences are:    Activate,  Read,   Precharge
#   2. Write sequences are:   Activate,  Write,  Precharge
#   3. Refresh sequences are: Precharge, Refresh
#   4. NOP seqences have a minimum length of 6 and do not use the 
#      complete signal.
#   5. The complete signal must be asserted on the 6th to last element of
#      a sequence.
#
# DDR Read Rules:
#   1. Activate to Read has minimum length of tRCD.
#   2. Read to Precharge has a minimum length of num_data_beats.
#   3. Activate to Precharge has a minimum length of tRAS, maximum length of tRASmax.
#   4. Precharge to Activate has a minimum length of tRP.
#   5. Activate to Activate has a minimum length of max(tRC, tRRD).
#
# DDR Write Rules:
#   1. Activate to Write has minimum length of tRCD.
#   2. Write to Precharge has a minimum length of tDQSS + num_data_beats + tWR.
#   3. Activate to Precharge has a minimum length of tRAS, maximum length of tRASmax.
#   4. Precharge to Activate has a minimum length of tRP.
#   5. Activate to Activate has a minimum length of max(tRC, tRRD).
#
# DDR Refresh Rules:
#   1. Precharge to Refresh has a minimum length of tRP.
#   2. Refresh to Activate has a minimum length of tRFC.
#
# DDR2 Read Rules:
#   1. Activate to Read has minimum length of tRCD - tAL.
#   2. Read to Precharge has a minimum length of 
#      tAL + burst_length/2 + tRTP - 2 clocks.
#   3. Activate to Precharge has a minimum length of tRAS, maximum length of tRASmax.
#   4. Precharge to Activate has a minimum length of tRPA.
#   5. Activate to Activate has a minimum length of max(tRC, tRRD).
#
# DDR2 Write Rules:
#   1. Activate to Write has minimum length of tRCD.
#   2. Write to Precharge has a minimum length of 
#      WL + max(burst_length/2,num_data_beats) + tWR.
#   3. Activate to Precharge has a minimum length of tRAS, maximum length of tRASmax.
#   4. Precharge to Activate has a minimum length of tRPA.
#   5. Activate to Activate has a minimum length of max(tRC, tRRD).
#
# DDR2 Refresh Rules:
#   1. Precharge to Refresh has a minimum length of tRPA.
#   2. Refresh to Activate has a minimum length of tRFC.
#
#
#
#--------------------------------------------------------------------------
#
# Sequence  0: Word Write
# Sequence  1: Word Read
# Sequence  2: CL4 Write
# Sequence  3: CL4 Read
# Sequence  4: CL8 Write
# Sequence  5: CL8 Read
# Sequence  6: B32 Write
# Sequence  7: B32 Read
# Sequence  8: B64 Write -- Not supported for 8-bit memories
# Sequence  9: B64 Read -- Not supported for 8-bit memories
# Sequence 10: NOP
# Sequence 11: NOP
# Sequence 12: NOP
# Sequence 13: NOP
# Sequence 14: Refresh
# Sequence 15: NOP
#
#--------------------------------------------------------------------------
#
# Control Signals:
#       0 : Ctrl_Complete
#       1 : Ctrl_Stall
#       2 : Ctrl_PhyIF_RAS_n
#       3 : Ctrl_PhyIF_CAS_n
#       4 : Ctrl_PhyIF_WE_n
#       5 : Ctrl_PhyIF_CE
#       6 : Ctrl_RMW         
#       7 : Ctrl_Skip[0]         
#       8 : Ctrl_PhyIF_DQS_O
#       9 : Ctrl_Skip[1]        
#      10 : Ctrl_DP_RdFIFO_Push
#      11 : Ctrl_Skip[2]         
#      12 : Ctrl_AP_Col_Cnt_Load
#      13 : Ctrl_AP_Col_Cnt_Enable
#      14 : Ctrl_AP_Precharge_Addr10
#      15 : Ctrl_AP_Row_Col_Sel
#      16 : Ctrl_Is_Write
#   31:17 : Unused
#      32 : Ctrl_PhyIF_Force_DM
#      33 : Ctrl_Repeat4
#   35:34 : Unused
#
#--------------------------------------------------------------------------

sub max {
    my $val0 = $_[0];
    my $val1 = $_[1];
    my $result = 0;
    $result = ($val0 > $val1) ? $val0 : $val1;
    return ($result);
}

sub bin2dec {
   my $temp = $_[0];
   my $result = 0;
   my $addfactor = 1;
   while ($temp ne "") {
      if (substr($temp,length($temp)-1,1)) { $result += $addfactor; }
      $addfactor *= 2;
      $temp = substr($temp,0,length($temp)-1);
   }
   return ($result);
}

sub dec2bin {
   my $temp = $_[0];
   my $result = "";
   my $divfactor = 32768;
   foreach my $k (1 .. 16) {
      if (int($temp/$divfactor)) { $result = $result."1"; $temp -= $divfactor; }
      else                       { $result = $result."0"; }
      $divfactor /= 2;
   }
   return ($result);
}

#### Subroutine for SDRAM STARTS HERE ###########

sub sdr_activate_to_read {
    my $tRCD   = $_[0];
    my $skip_0 = $_[1];
    my $skip_1 = $_[2];
    my $skip_2 = $_[3];
    my $family = $_[4];
    my @table;
    my $i;
    my $j;
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]    
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        if ($i == ceil($tRCD) - 2) {
            $table[$j] = $table[$j] . $skip_0;
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        if ($i == ceil($tRCD) - 2) {
            $table[$j] = $table[$j] . $skip_1;
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        if ($i == ceil($tRCD) - 2) {
            $table[$j] = $table[$j] . $skip_2;
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Return Table
    @table;
}


sub sdr_read_to_precharge {
    my $num_data_beats = $_[0];
    my $burst_length   = $_[1];
    my $rmw            = $_[2];
    my $allow_repeat   = $_[3];
    my $family         = $_[4];
    my @table;
    my $i;
    my $j;
    my $num_data_beats_tmp;
    if (($rmw == 1) && ($num_data_beats < $burst_length)) {
        $num_data_beats = $burst_length;
    }
    if ($num_data_beats < $burst_length) {
        $num_data_beats_tmp = $burst_length;
    }
    else {
        $num_data_beats_tmp = $num_data_beats;
    }
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }
    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if ((( $i % ($burst_length) ) == 0) && ($i < $num_data_beats_tmp)) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if ($i<$num_data_beats) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    $table[$j] = $table[$j] . '1';
    for ($i=1;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if ($i==0) {
            $table[$j] = $table[$j] . '0';
        }
        elsif ((( $i % ($burst_length) ) == 0) && ($i < $num_data_beats_tmp)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$num_data_beats_tmp;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$num_data_beats_tmp;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if (($i<$num_data_beats_tmp) && ($i>=4) && ($i%4==0) && 
            ($allow_repeat==1)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$num_data_beats_tmp;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Return Table
    @table;
}

sub sdr_activate_to_read_to_precharge {
    my $tRCD           = $_[0];
    my $num_data_beats = $_[1];
    my $burst_length   = $_[2];
    my $tRAS           = $_[3];
    my $tRASmax        = $_[4];
    my $allow_repeat   = $_[5];
    my $family         = $_[6];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;
    
    ###Get table for activate to read
    @table = &sdr_activate_to_read($tRCD,0,0,0,$family);

    ###Append table for read to precharge
    @table_tmp = &sdr_read_to_precharge($num_data_beats,$burst_length,0,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    if ($tRASmax < $table_length) {
            print "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.  For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
            print ERRORFILE "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
            die print "Aborting script: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
    }
    @table_tmp = &sdr_nop($tRAS-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    # Return Table
    @table;
}

sub sdr_activate_to_read_to_write_to_precharge {
    my $tRCD           = $_[0];
    my $tDQSS          = $_[1];
    my $num_data_beats = $_[2];
    my $tWR            = $_[3];
    my $burst_length   = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $cas_latency    = $_[7];
    my $skip_0         = $_[8];
    my $skip_1         = $_[9];
    my $skip_2         = $_[10];
    my $allow_repeat   = $_[11];
    my $family         = $_[12];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;
    my $tmp;
    
    ###Get table for activate to read
    @table = &sdr_activate_to_read($tRCD,$skip_0,$skip_1,$skip_2,$family);

    ###Append table for read to precharge
    @table_tmp = &sdr_read_to_precharge($num_data_beats,$burst_length,1,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table_tmp[0]);
    if ($num_data_beats > $burst_length) {
        $tmp = $num_data_beats + $cas_latency;
    }
    else {
        $tmp = $burst_length + $cas_latency;
    }
    if ($tmp < $burst_length + $cas_latency + 9) {
        $tmp = $burst_length + $cas_latency + 9;
    }
    if ($table_length < $tmp) {
        @table_tmp = &sdr_nop($tmp - $table_length);
        for ($i=0;$i<36;$i++) {
            $table[$i] = $table[$i] . $table_tmp[$i];
        }
    }

    ###Append table for write to precharge
    @table_tmp = &sdr_write_to_precharge(0,$num_data_beats,$tWR,$burst_length,1,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    if ($tRASmax < $table_length) {
            print "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.  For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
            print ERRORFILE "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
            die print "Aborting script: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
    }
    @table_tmp = &sdr_nop($tRAS-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    # Return Table
    @table;
}

sub sdr_activate_to_read_to_write_to_precharge_to_activate {
    my $tRCD           = $_[0];
    my $tDQSS          = $_[1];
    my $num_data_beats = $_[2];
    my $tWR            = $_[3];
    my $burst_length   = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $tRP            = $_[7];
    my $tRC            = $_[8];
    my $tRRD           = $_[9];
    my $cas_latency    = $_[10];
    my $skip_0         = $_[11];
    my $skip_1         = $_[12];
    my $skip_2         = $_[13];
    my $allow_repeat   = $_[14];
    my $family         = $_[15];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for activate to write to precharge
    @table = &sdr_activate_to_read_to_write_to_precharge($tRCD,0,$num_data_beats,$tWR,$burst_length,$tRAS,$tRASmax,$cas_latency,$skip_0,$skip_1,$skip_2,$allow_repeat,$family);

    ###Append table for precharge to activate
    @table_tmp = &sdr_precharge_to_activate($tRP,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    @table_tmp = &sdr_nop(&max($tRC, $tRRD)-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &sdr_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    ### Overwrite Ctrl_RMW Bit
    $table_length = length($table[0]);
    $table[6] = '';
    for ($i=0;$i<$table_length;$i++) {

sub sdr_precharge_to_activate {
    my $tRP    = $_[0];
    my $family = $_[1];
    my @table;
    my $i;
    my $j;
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]    
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    $table[$j] = $table[$j] . '1';
    for ($i=1;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$tRP;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$tRP;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$tRP;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Return Table
    @table;
}

        $table[6] = $table[6] . '1';
    }
    

    # Return Table
    @table;
}

sub sdr_activate_to_read_to_precharge_to_activate {
    my $tRCD           = $_[0];
    my $num_data_beats = $_[1];
    my $burst_length   = $_[2];
    my $tRAS           = $_[3];
    my $tRASmax        = $_[4];
    my $tRP            = $_[5];
    my $tRC            = $_[6];
    my $tRRD           = $_[7];
    my $allow_repeat   = $_[8];
    my $family         = $_[9];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for activate to read to precharge
    @table = &sdr_activate_to_read_to_precharge($tRCD,$num_data_beats,$burst_length,$tRAS,$tRASmax,$allow_repeat,$family);

    ###Append table for precharge to activate
    @table_tmp = &sdr_precharge_to_activate($tRP,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    @table_tmp = &sdr_nop($tRC-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &sdr_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    # Return Table
    @table;
}

sub sdr_activate_to_write {
    my $tRCD   = $_[0];
    my $family = $_[1];
    my @table= ();
    my $i;
    my $j;

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }
    $j=0;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]    
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }
    
    # Return Table
    @table;
}

sub sdr_write_to_precharge {
    my $tDQSS          = $_[0];
    my $num_data_beats = $_[1];
    my $tWR            = $_[2];
    my $burst_length   = $_[3];
    my $rmw            = $_[4];
    my $allow_repeat   = $_[5];
    my $family         = $_[6];
    my @table;
    my @table_tmp;
    my $i;
    my $j;
    my $num_data_beats_tmp;
    my $sequence_length;

    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }
    if ($num_data_beats < $burst_length) {
        $num_data_beats_tmp = $burst_length;
    }
    else {
        $num_data_beats_tmp = $num_data_beats;
    }

    if ($num_data_beats+$tWR < $burst_length) {
        $sequence_length = $burst_length;
    }
    else {
        $sequence_length = $num_data_beats+$tWR;
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        if ((( $i % ($burst_length) ) == 0) && ($i < $num_data_beats)) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        if ((( $i % ($burst_length) ) == 0) && ($i < $num_data_beats)) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        if ($i<$num_data_beats_tmp) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        if ($i==0) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        if ($i==0) {
            $table[$j] = $table[$j] . '0';
        }
        elsif ((($i % ($burst_length)) == 0) && ($i < $num_data_beats)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$sequence_length;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$sequence_length;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$sequence_length;$i++) {
        if ($i<$num_data_beats) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$sequence_length;$i++) {
        if (($i<$num_data_beats_tmp) && ($i>=4) && ($i%4==0) && 
            ($allow_repeat==1)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$sequence_length;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    if (($rmw == 1) && ($num_data_beats < $burst_length)) {
        @table_tmp = &sdr_nop($burst_length - $num_data_beats);
        for ($i=0;$i<36;$i++) {
            $table[$i] = $table[$i] . $table_tmp[$i];
        }
    }

    # Return Table
    @table;
}

sub sdr_activate_to_write_to_precharge {
    my $tRCD           = $_[0];
    my $tDQSS          = $_[1];
    my $num_data_beats = $_[2];
    my $tWR            = $_[3];
    my $burst_length   = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $allow_repeat   = $_[7];
    my $family         = $_[8];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for activate to write
    @table = &sdr_activate_to_write($tRCD,$family);

    ###Append table for write to precharge
    @table_tmp = &sdr_write_to_precharge(0,$num_data_beats,$tWR,$burst_length,0,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    @table_tmp = &sdr_nop($tRAS-$table_length);
    if ($tRASmax < $table_length) {
            print "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.  For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
            print ERRORFILE "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
            die "Aborting script: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
    }
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    # Return Table
    @table;
}

sub sdr_activate_to_write_to_precharge_to_activate {
    my $tRCD           = $_[0];
    my $tDQSS          = $_[1];
    my $num_data_beats = $_[2];
    my $tWR            = $_[3];
    my $burst_length   = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $tRP            = $_[7];
    my $tRC            = $_[8];
    my $tRRD           = $_[9];
    my $allow_repeat   = $_[10];
    my $family         = $_[11];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for activate to write to precharge
    @table = &sdr_activate_to_write_to_precharge($tRCD,0,$num_data_beats,$tWR,$burst_length,$tRAS,$tRASmax,$allow_repeat,$family);

    ###Append table for precharge to activate
    @table_tmp = &sdr_precharge_to_activate($tRP,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    @table_tmp = &sdr_nop(&max($tRC, $tRRD)-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &sdr_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    # Return Table
    @table;
}

sub sdr_precharge_to_refresh {
    my $tRP    = $_[0];
    my $family = $_[1];
    my @table;
    my $i;
    my $j;

    @table = &sdr_precharge_to_activate($tRP,$family);

    # Return Table
    @table;
}

sub sdr_refresh_to_activate {
    my $tRFC   = $_[0];
    my $family = $_[1];
    my @table;
    my $i;
    my $j;
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$tRFC;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$tRFC;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$tRFC;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Return Table
    @table;
}

sub sdr_precharge_to_refresh_to_activate {
    my $tRP    = $_[0];
    my $tRFC   = $_[1];
    my $family = $_[2];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for precharge to refresh
    @table = &sdr_precharge_to_refresh($tRP,$family);

    ###Append table for refresh to activate
    @table_tmp = &sdr_refresh_to_activate($tRFC,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &sdr_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    # Return Table
    @table;    
}

sub sdr_nop {
    my $length = $_[0];
    my @table;
    my $i;
    my $j;
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$length;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$length;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$length;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Return Table
    @table;
}


#### Subroutine for SDRAM ENDS HERE #############

sub ddr_activate_to_read {
    my $tRCD   = $_[0];
    my $skip_0 = $_[1];
    my $skip_1 = $_[2];
    my $skip_2 = $_[3];
    my $family = $_[4];
    my @table;
    my $i;
    my $j;
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]    
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        if ($i == ceil($tRCD) - 2) {
            $table[$j] = $table[$j] . $skip_0;
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        if ($i == ceil($tRCD) - 2) {
            $table[$j] = $table[$j] . $skip_1;
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        if ($i == ceil($tRCD) - 2) {
            $table[$j] = $table[$j] . $skip_2;
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Return Table
    @table;
}

sub ddr_read_to_precharge {
    my $num_data_beats = $_[0];
    my $burst_length   = $_[1];
    my $rmw            = $_[2];
    my $allow_repeat   = $_[3];
    my $family         = $_[4];
    my @table;
    my $i;
    my $j;
    my $num_data_beats_tmp;
    if (($rmw == 1) && ($num_data_beats < $burst_length/2)) {
        $num_data_beats = $burst_length/2;
    }
    if ($num_data_beats < $burst_length/2) {
        $num_data_beats_tmp = $burst_length/2;
    }
    else {
        $num_data_beats_tmp = $num_data_beats;
    }
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }
    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if ((( $i % ($burst_length/2) ) == 0) && ($i < $num_data_beats_tmp)) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if ($i<$num_data_beats) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    $table[$j] = $table[$j] . '1';
    for ($i=1;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if ($i==0) {
            $table[$j] = $table[$j] . '0';
        }
        elsif ((( $i % ($burst_length/2) ) == 0) && ($i < $num_data_beats_tmp)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$num_data_beats_tmp;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$num_data_beats_tmp;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if (($i<$num_data_beats_tmp) && ($i>=4) && ($i%4==0) && 
            ($allow_repeat==1)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$num_data_beats_tmp;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Return Table
    @table;
}

sub ddr_activate_to_read_to_precharge {
    my $tRCD           = $_[0];
    my $num_data_beats = $_[1];
    my $burst_length   = $_[2];
    my $tRAS           = $_[3];
    my $tRASmax        = $_[4];
    my $allow_repeat   = $_[5];
    my $family         = $_[6];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;
    
    ###Get table for activate to read
    @table = &ddr_activate_to_read($tRCD,0,0,0,$family);

    ###Append table for read to precharge
    @table_tmp = &ddr_read_to_precharge($num_data_beats,$burst_length,0,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    if ($tRASmax < $table_length) {
            print "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.  For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
            print ERRORFILE "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
            die print "Aborting script: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
    }
    @table_tmp = &ddr_nop($tRAS-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    # Return Table
    @table;
}

sub ddr_activate_to_read_to_write_to_precharge {
    my $tRCD           = $_[0];
    my $tDQSS          = $_[1];
    my $num_data_beats = $_[2];
    my $tWR            = $_[3];
    my $burst_length   = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $cas_latency    = $_[7];
    my $skip_0         = $_[8];
    my $skip_1         = $_[9];
    my $skip_2         = $_[10];
    my $allow_repeat   = $_[11];
    my $family         = $_[12];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;
    my $tmp;
    
    ###Get table for activate to read
    @table = &ddr_activate_to_read($tRCD,$skip_0,$skip_1,$skip_2,$family);
    ###Special Case for Spartan3 ECC.  Adds a bubble cycle.
    if ($family eq 'spartan3') {
        @table_tmp = &ddr_nop(1);
        for ($i=0;$i<36;$i++) {
            $table[$i] = $table[$i] . $table_tmp[$i];
        }
    }

    ###Append table for read to precharge
    @table_tmp = &ddr_read_to_precharge($num_data_beats,$burst_length,1,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table_tmp[0]);
    if ($num_data_beats > $burst_length/2) {
        $tmp = $num_data_beats + $cas_latency;
    }
    else {
        $tmp = $burst_length/2 + $cas_latency;
    }
    if ($tmp < $burst_length/2 + $cas_latency + 10) {
        $tmp = $burst_length/2 + $cas_latency + 10;
    }
    if ($family eq 'spartan3') {
        $tmp = $tmp + 6;
    }
    if ($table_length < $tmp) {
        @table_tmp = &ddr_nop($tmp - $table_length);
        for ($i=0;$i<36;$i++) {
            $table[$i] = $table[$i] . $table_tmp[$i];
        }
    }

    ###Append table for write to precharge
    @table_tmp = &ddr_write_to_precharge($tDQSS,$num_data_beats,$tWR,$burst_length,1,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    if ($tRASmax < $table_length) {
            print "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.  For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
            print ERRORFILE "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
            die print "Aborting script: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
    }
    @table_tmp = &ddr_nop($tRAS-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    # Return Table
    @table;
}

sub ddr_activate_to_read_to_write_to_precharge_to_activate {
    my $tRCD           = $_[0];
    my $tDQSS          = $_[1];
    my $num_data_beats = $_[2];
    my $tWR            = $_[3];
    my $burst_length   = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $tRP            = $_[7];
    my $tRC            = $_[8];
    my $tRRD           = $_[9];
    my $cas_latency    = $_[10];
    my $skip_0         = $_[11];
    my $skip_1         = $_[12];
    my $skip_2         = $_[13];
    my $allow_repeat   = $_[14];
    my $family         = $_[15];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for activate to write to precharge
    @table = &ddr_activate_to_read_to_write_to_precharge($tRCD,$tDQSS,$num_data_beats,$tWR,$burst_length,$tRAS,$tRASmax,$cas_latency,$skip_0,$skip_1,$skip_2,$allow_repeat,$family);

    ###Append table for precharge to activate
    @table_tmp = &ddr_precharge_to_activate($tRP,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    @table_tmp = &ddr_nop(&max($tRC, $tRRD)-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &ddr_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    ### Overwrite Ctrl_RMW Bit
    $table_length = length($table[0]);
    $table[6] = '';
    $table[6] = $table[6] . '0';
    for ($i=1;$i<$table_length;$i++) {
        $table[6] = $table[6] . '1';
    }
    

    # Return Table
    @table;
}

sub ddr_precharge_to_activate {
    my $tRP    = $_[0];
    my $family = $_[1];
    my @table;
    my $i;
    my $j;
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]    
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    $table[$j] = $table[$j] . '1';
    for ($i=1;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$tRP;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$tRP;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$tRP;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$tRP;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Return Table
    @table;
}

sub ddr_activate_to_read_to_precharge_to_activate {
    my $tRCD           = $_[0];
    my $num_data_beats = $_[1];
    my $burst_length   = $_[2];
    my $tRAS           = $_[3];
    my $tRASmax        = $_[4];
    my $tRP            = $_[5];
    my $tRC            = $_[6];
    my $tRRD           = $_[7];
    my $allow_repeat   = $_[8];
    my $family         = $_[9];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for activate to read to precharge
    @table = &ddr_activate_to_read_to_precharge($tRCD,$num_data_beats,$burst_length,$tRAS,$tRASmax,$allow_repeat,$family);

    ###Append table for precharge to activate
    @table_tmp = &ddr_precharge_to_activate($tRP,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    @table_tmp = &ddr_nop($tRC-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &ddr_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    # Return Table
    @table;
}

sub ddr_activate_to_write {
    my $tRCD   = $_[0];
    my $family = $_[1];
    my @table= ();
    my $i;
    my $j;

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }
    $j=0;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]    
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }
    
    # Return Table
    @table;
}

sub ddr_write_to_precharge {
    my $tDQSS          = $_[0];
    my $num_data_beats = $_[1];
    my $tWR            = $_[2];
    my $burst_length   = $_[3];
    my $rmw            = $_[4];
    my $allow_repeat   = $_[5];
    my $family         = $_[6];
    my @table;
    my @table_tmp;
    my $i;
    my $j;
    my $num_data_beats_tmp;

    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }
    if ($num_data_beats < $burst_length/2) {
        $num_data_beats_tmp = $burst_length/2;
    }
    else {
        $num_data_beats_tmp = $num_data_beats;
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        if ((( $i % ($burst_length/2) ) == 0) && ($i < $num_data_beats)) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        if ((( $i % ($burst_length/2) ) == 0) && ($i < $num_data_beats)) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        if ($i<$num_data_beats_tmp) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        if ($i==0) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        if ($i==0) {
            $table[$j] = $table[$j] . '0';
        }
        elsif ((($i % ($burst_length/2)) == 0) && ($i < $num_data_beats)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        if ($i<$num_data_beats) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
        if (($i<$num_data_beats_tmp) && ($i>=4) && ($i%4==0) && 
            ($allow_repeat==1)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$tDQSS+$num_data_beats+$tWR;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    if (($rmw == 1) && ($num_data_beats < $burst_length/2)) {
        @table_tmp = &ddr_nop($burst_length/2 - $num_data_beats);
        for ($i=0;$i<36;$i++) {
            $table[$i] = $table[$i] . $table_tmp[$i];
        }
    }

    # Return Table
    @table;
}

sub ddr_activate_to_write_to_precharge {
    my $tRCD           = $_[0];
    my $tDQSS          = $_[1];
    my $num_data_beats = $_[2];
    my $tWR            = $_[3];
    my $burst_length   = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $allow_repeat   = $_[7];
    my $family         = $_[8];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for activate to write
    @table = &ddr_activate_to_write($tRCD,$family);

    ###Append table for write to precharge
    @table_tmp = &ddr_write_to_precharge($tDQSS,$num_data_beats,$tWR,$burst_length,0,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    @table_tmp = &ddr_nop($tRAS-$table_length);
    if ($tRASmax < $table_length) {
            print "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.  For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
            print ERRORFILE "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
            die "Aborting script: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
    }
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    # Return Table
    @table;
}

sub ddr_activate_to_write_to_precharge_to_activate {
    my $tRCD           = $_[0];
    my $tDQSS          = $_[1];
    my $num_data_beats = $_[2];
    my $tWR            = $_[3];
    my $burst_length   = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $tRP            = $_[7];
    my $tRC            = $_[8];
    my $tRRD           = $_[9];
    my $allow_repeat   = $_[10];
    my $family         = $_[11];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for activate to write to precharge
    @table = &ddr_activate_to_write_to_precharge($tRCD,$tDQSS,$num_data_beats,$tWR,$burst_length,$tRAS,$tRASmax,$allow_repeat,$family);

    ###Append table for precharge to activate
    @table_tmp = &ddr_precharge_to_activate($tRP,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    @table_tmp = &ddr_nop(&max($tRC, $tRRD)-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &ddr_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    # Return Table
    @table;
}

sub ddr_precharge_to_refresh {
    my $tRP    = $_[0];
    my $family = $_[1];
    my @table;
    my $i;
    my $j;

    @table = &ddr_precharge_to_activate($tRP,$family);

    # Return Table
    @table;
}

sub ddr_refresh_to_activate {
    my $tRFC   = $_[0];
    my $family = $_[1];
    my @table;
    my $i;
    my $j;
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$tRFC;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$tRFC;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$tRFC;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$tRFC;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Return Table
    @table;
}

sub ddr2_refresh_to_activate {
    my $tRFC   = $_[0];
    my $family = $_[1];
    my @table;
    my $i;
    my $j;

    @table = &ddr_refresh_to_activate($tRFC,$family);

    # Return Table
    @table;
}

sub ddr_precharge_to_refresh_to_activate {
    my $tRP    = $_[0];
    my $tRFC   = $_[1];
    my $family = $_[2];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for precharge to refresh
    @table = &ddr_precharge_to_refresh($tRP,$family);

    ###Append table for refresh to activate
    @table_tmp = &ddr_refresh_to_activate($tRFC,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &ddr_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    # Return Table
    @table;    
}

sub ddr_nop {
    my $length = $_[0];
    my @table;
    my $i;
    my $j;
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$length;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$length;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$length;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$length;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Return Table
    @table;
}

sub ddr2_activate_to_read {
    my $tRCD   = $_[0];
    my $tAL    = $_[1];
    my $skip_0 = $_[2];
    my $skip_1 = $_[3];
    my $skip_2 = $_[4];
    my $family = $_[5];
    my @table;
    my $i;
    my $j;

    @table = &ddr_activate_to_read($tRCD - $tAL,$skip_0,$skip_1,$skip_2,$family);

    # Return Table
    @table;
}

sub ddr2_read_to_precharge {
    my $num_data_beats = $_[0];
    my $burst_length   = $_[1];
    my $tAL            = $_[2];
    my $tRTP           = $_[3];
    my $rmw            = $_[4];
    my $allow_repeat   = $_[5];
    my $family         = $_[6];
    my @table;
    my @table_tmp;
    my $num_data_beats_tmp;
    my $i;
    my $j;
    if (($rmw == 1) && ($num_data_beats < $burst_length/2)) {
        $num_data_beats = $burst_length/2;
    }
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }
    if ($num_data_beats < $burst_length/2) {
        $num_data_beats_tmp = $burst_length/2;
    }
    else {
        $num_data_beats_tmp = $num_data_beats;
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if ((( $i % ($burst_length/2) ) == 0) && ($i < $num_data_beats)) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if ($i<$num_data_beats) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    $table[$j] = $table[$j] . '1';
    for ($i=1;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if ($i==0) {
            $table[$j] = $table[$j] . '0';
        }
        elsif ((( $i % ($burst_length/2) ) == 0) && ($i < $num_data_beats)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$num_data_beats_tmp;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$num_data_beats_tmp;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$num_data_beats_tmp;$i++) {
        if (($i<$num_data_beats_tmp) && ($i>=4) && ($i%4==0) && 
            ($allow_repeat==1)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$num_data_beats_tmp;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    ### Add NOPs to table if necessary
    if ($tAL+$tRTP-2>0) {
        @table_tmp = &ddr2_nop($tAL+$tRTP-2);
        for ($i=0;$i<36;$i++) {
            $table[$i] = $table[$i] . $table_tmp[$i];
        }
    }

    # Return Table
    @table;
}

sub ddr2_activate_to_read_to_precharge {
    my $tRCD           = $_[0];
    my $tAL            = $_[1];
    my $num_data_beats = $_[2];
    my $burst_length   = $_[3];
    my $tRTP           = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $allow_repeat   = $_[7];
    my $family         = $_[8];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;
    
    ###Get table for activate to read
    @table = &ddr2_activate_to_read($tRCD,$tAL,0,0,0,$family);

    ###Append table for read to precharge
    @table_tmp = &ddr2_read_to_precharge($num_data_beats,$burst_length,$tAL,$tRTP,0,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    if ($tRASmax < $table_length) {
            print "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.  For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
            print ERRORFILE "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
            die "Aborting script: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
    }
    @table_tmp = &ddr2_nop($tRAS-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    # Return Table
    @table;
}

sub ddr2_precharge_to_activate {
    my $tRPA   = $_[0];
    my $family = $_[1];
    my @table;
    my $i;
    my $j;

    @table = &ddr_precharge_to_activate($tRPA,$family);

    # Return Table
    @table;
}

sub ddr2_activate_to_read_to_precharge_to_activate {
    my $tRCD           = $_[0];
    my $tAL            = $_[1];
    my $num_data_beats = $_[2];
    my $burst_length   = $_[3];
    my $tRTP           = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $tRPA           = $_[7];
    my $tRC            = $_[8];
    my $tRRD           = $_[9];
    my $allow_repeat   = $_[10];
    my $family         = $_[11];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for activate to read to precharge
    @table = &ddr2_activate_to_read_to_precharge($tRCD,$tAL,$num_data_beats,$burst_length,$tRTP,$tRAS,$tRASmax,$allow_repeat,$family);
    
    ###Append table for precharge to activate
    @table_tmp = &ddr2_precharge_to_activate($tRPA,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    @table_tmp = &ddr2_nop(&max($tRC, $tRRD)-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &ddr2_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    # Return Table
    @table;
}

sub ddr2_activate_to_read_to_write_to_precharge {
    my $tRCD           = $_[0];
    my $tAL            = $_[1];
    my $num_data_beats = $_[2];
    my $burst_length   = $_[3];
    my $tRTP           = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $tWL            = $_[7];
    my $tWR            = $_[8];
    my $ODT_Rtt        = $_[9];
    my $cas_latency    = $_[10];
    my $skip_0         = $_[11];
    my $skip_1         = $_[12];
    my $skip_2         = $_[13];
    my $allow_repeat   = $_[14];
    my $family         = $_[15];
    my @table;
    my @table_tmp;
    my $table_length;
    my $tmp;
    my $i;
    my $j;
    
    ###Get table for activate to read
    @table = &ddr2_activate_to_read($tRCD,$tAL,$skip_0,$skip_1,$skip_2,$family);
    ###Special Case for Spartan3 ECC.  Adds a bubble cycle.
    if ($family eq 'spartan3') {
        @table_tmp = &ddr2_nop(1);
        for ($i=0;$i<36;$i++) {
            $table[$i] = $table[$i] . $table_tmp[$i];
        }
    }

    ###Append table for read to precharge
    @table_tmp = &ddr2_read_to_precharge($num_data_beats,$burst_length,$tAL,2,1,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table_tmp[0]);
    if ($num_data_beats > $burst_length/2) {
        $tmp = $num_data_beats + $tAL + $cas_latency;
    }
    else {
        $tmp = $burst_length/2 + $tAL + $cas_latency;
    }
    if ($tmp < $burst_length/2 + $tAL + $cas_latency + 9) {
        $tmp = $burst_length/2 + $tAL + $cas_latency + 9;
    }
    if ($family eq 'spartan3') {
        $tmp = $tmp + 6;
    }
    if ($table_length < $tmp) {
        @table_tmp = &ddr2_nop($tmp - $table_length);
        for ($i=0;$i<36;$i++) {
            $table[$i] = $table[$i] . $table_tmp[$i];
        }
    }

    ###Append table for write to precharge
    @table_tmp = &ddr2_write_to_precharge($tWL,$num_data_beats,$tWR,$burst_length,$ODT_Rtt,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    if ($tRASmax < $table_length) {
            print "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.  For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
            print ERRORFILE "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
            die "Aborting script: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
    }
    @table_tmp = &ddr2_nop($tRAS-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    # Return Table
    @table;
}

sub ddr2_activate_to_read_to_write_to_precharge_to_activate {
    my $tRCD           = $_[0];
    my $tAL            = $_[1];
    my $num_data_beats = $_[2];
    my $burst_length   = $_[3];
    my $tRTP           = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $tRPA           = $_[7];
    my $tRC            = $_[8];
    my $tRRD           = $_[9];
    my $tWL            = $_[10];
    my $tWR            = $_[11];
    my $ODT_Rtt        = $_[12];
    my $cas_latency    = $_[13];
    my $skip_0         = $_[14];
    my $skip_1         = $_[15];
    my $skip_2         = $_[16];
    my $allow_repeat   = $_[17];
    my $family         = $_[18];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for activate to read to precharge
    @table = &ddr2_activate_to_read_to_write_to_precharge($tRCD,$tAL,$num_data_beats,$burst_length,$tRTP,$tRAS,$tRASmax,$tWL,$tWR,$ODT_Rtt,$cas_latency,$skip_0,$skip_1,$skip_2,$allow_repeat,$family);
    
    ###Append table for precharge to activate
    @table_tmp = &ddr2_precharge_to_activate($tRPA,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    @table_tmp = &ddr2_nop(&max($tRC, $tRRD)-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &ddr2_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    ### Overwrite Ctrl_RMW Bit
    $table_length = length($table[0]);
    $table[6] = '';
    for ($i=0;$i<$table_length;$i++) {
        $table[6] = $table[6] . '1';
    }

    # Return Table
    @table;
}

sub ddr2_activate_to_write {
    my $tRCD       = $_[0];
    my $ODT_Rtt    = $_[1];
    my $family     = $_[2];
    my @table;
    my $i;
    my $j;

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }
    $j=0;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]    
    $j++;
    $table[$j] = $table[$j] . '0';
    for ($i=1;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$tRCD;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$tRCD;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }
    
    # Return Table
    @table;
}

sub ddr2_write_to_precharge {
    my $tWL            = $_[0];
    my $num_data_beats = $_[1];
    my $tWR            = $_[2];
    my $burst_length   = $_[3];
    my $ODT_Rtt        = $_[4];
    my $allow_repeat   = $_[5];
    my $family         = $_[6];
    my @table;
    my $i;
    my $j;
    my $num_data_beats_tmp;
    
    for ($i=0;$i<36;$i++) {
        $table[$i] = '';
    }

    if ($num_data_beats < $burst_length/2) {
        $num_data_beats_tmp = $burst_length/2;
    }
    else {
        $num_data_beats_tmp = $num_data_beats;
    }

    # Ctrl_Complete:             $table[0], set to 0, may be overwritten.
    $j=0;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Stall:                $table[1]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_RAS_n:          $table[2]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_PhyIF_CAS_n:          $table[3]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        if ((( $i % ($burst_length/2) ) == 0) && ($i < $num_data_beats)) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_PhyIF_WE_n:           $table[4]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        if ((( $i % ($burst_length/2) ) == 0) && ($i < $num_data_beats)) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_PhyIF_CE:             $table[5]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '1';
    }

    # Ctrl_RMW:                  $table[6]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[0]               $table[7]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_PhyIF_DQS_O:          $table[8]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        if ($i<$num_data_beats_tmp) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_Skip[1]               $table[9]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_DP_RdFIFO_Push:       $table[10]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Skip[2]               $table[11]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Col_Cnt_Load:      $table[12]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        if ($i==0) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_AP_Col_Cnt_Enable:    $table[13]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        if ($i==0) {
            $table[$j] = $table[$j] . '0';
        }
        elsif ((($i % ($burst_length/2)) == 0) && ($i < $num_data_beats)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_AP_Precharge_Addr10:  $table[14]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_AP_Row_Col_Sel:       $table[15]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Ctrl_Is_Write:             $table[16]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        $table[$j] = $table[$j] . '0';
    }

    # Unused:                    $table[29:17]
    for ($j=17;$j<30;$j++) {
        for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:                    $table[31:30]
    for ($j=30;$j<32;$j++) {
        for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Ctrl_PhyIF_Force_DM:       $table[32]
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        if ($i<$num_data_beats) {
            $table[$j] = $table[$j] . '0';
        }
        else {
            $table[$j] = $table[$j] . '1';
        }
    }

    # Ctrl_Repeat4:              $table[33]
    $j++;
    for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
        if (($i<$num_data_beats_tmp) && ($i>=4) && ($i%4==0) && 
            ($allow_repeat==1)) {
            $table[$j] = $table[$j] . '1';
        }
        else {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Unused:     $table[35:34]
    for ($j=34;$j<36;$j++) {
        for ($i=0;$i<$tWL+$num_data_beats_tmp+$tWR;$i++) {
            $table[$j] = $table[$j] . '0';
        }
    }

    # Return Table
    @table;
}

sub ddr2_activate_to_write_to_precharge {
    my $tRCD           = $_[0];
    my $tWL            = $_[1];
    my $num_data_beats = $_[2];
    my $tWR            = $_[3];
    my $burst_length   = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $ODT_Rtt        = $_[7];
    my $allow_repeat   = $_[8];
    my $family         = $_[9];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;
    ###Get table for activate to write
    @table = &ddr2_activate_to_write($tRCD,$ODT_Rtt,$family);

    ###Append table for write to precharge
    @table_tmp = &ddr2_write_to_precharge($tWL,$num_data_beats,$tWR,$burst_length,$ODT_Rtt,$allow_repeat,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    if ($tRASmax < $table_length) {
            print "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.  For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
            print ERRORFILE "Error: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
            die "Aborting script: tRASmax value of $tRASmax has been violated by " . ($tRASmax-$table_length) . " cycles.\n";
    }
    @table_tmp = &ddr2_nop($tRAS-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    # Return Table
    @table;
}

sub ddr2_activate_to_write_to_precharge_to_activate {
    my $tRCD           = $_[0];
    my $tWL            = $_[1];
    my $num_data_beats = $_[2];
    my $tWR            = $_[3];
    my $burst_length   = $_[4];
    my $tRAS           = $_[5];
    my $tRASmax        = $_[6];
    my $tRPA           = $_[7];
    my $tRC            = $_[8];
    my $tRRD           = $_[9];
    my $ODT_Rtt        = $_[10];
    my $allow_repeat   = $_[11];
    my $family         = $_[12];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for activate to write to precharge
    @table = &ddr2_activate_to_write_to_precharge($tRCD,$tWL,$num_data_beats,$tWR,$burst_length,$tRAS,$tRASmax,$ODT_Rtt,$allow_repeat,$family);

    ###Append table for precharge to activate
    @table_tmp = &ddr2_precharge_to_activate($tRPA,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Add NOPs to table if necessary
    $table_length = length($table[0]);
    @table_tmp = &ddr2_nop(&max($tRC, $tRRD)-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &ddr2_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    # Return Table
    @table;
}

sub ddr2_precharge_to_refresh {
    my $tRPA   = $_[0];
    my $family = $_[1];
    my @table;
    my $i;
    my $j;

    @table = &ddr2_precharge_to_activate($tRPA,$family);

    # Return Table
    @table;
}

sub ddr2_precharge_to_refresh_to_activate {
    my $tRPA   = $_[0];
    my $tRFC   = $_[1];
    my $family = $_[2];
    my @table;
    my @table_tmp;
    my $table_length;
    my $i;
    my $j;

    ###Get table for precharge to refresh
    @table = &ddr2_precharge_to_refresh($tRPA,$family);

    ###Append table for refresh to activate
    @table_tmp = &ddr2_refresh_to_activate($tRFC,$family);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }

    ### Overwrite Complete Bit
    $table_length = length($table[0]);
    @table_tmp = &ddr2_nop(6-$table_length);
    for ($i=0;$i<36;$i++) {
        $table[$i] = $table[$i] . $table_tmp[$i];
    }
    $table_length = length($table[0]);
    $table[0] = '';
    for ($i=0;$i<$table_length-6;$i++) {
        $table[0] = $table[0] . '0';
    }
    $table[0] = $table[0] . '1';
    for ($i=0;$i<5;$i++) {
        $table[0] = $table[0] . '0';
    }

    # Return Table
    @table;    
}

sub ddr2_nop {
    my $length = $_[0];
    my @table;
    my $i;
    my $j;

    @table = &ddr_nop($length);

    # Return Table
    @table;
}

my $tRPA;
my $tAL;
my $tWL;
my $tRCD;
my $num_data_beats;
my $burst_length;
my $tRAS;
my $tRASmax;
my $tRP;
my $tRC;
my $tRRD;
my $tDQSS;
my $tWR;
my $tWR_ps;
my $tRFC;
my $tStallCkeLow;
my $tMRD;
my $cas_latency;
my $tCCD;
my $tWTR;           # currently unused
my @table;
my @table_tmp;
my $table_length;
my $i;
my $j;
my $k;
my $usage;
my $fout_txt;
my $fout_ver;
my $fout_err;
my $clk_freq;
my $reg_dimm;
my $data_width;
my $cp_pipeline;
my $ap_pipeline1;
my $ap_pipeline2;
my $ddr2;
my $tStallDllRst;
my $CAS_Latency_A_Fmax;
my $CAS_Latency_A;
my $CAS_Latency_B_Fmax;
my $CAS_Latency_B;
my $CAS_Latency_C_Fmax;
my $CAS_Latency_C;
my $CAS_Latency_D_Fmax;
my $CAS_Latency_D;
my $x4;
my $tRTP;
my $tAOND;
my $tAOFD;
my $tStallCkeHigh;
my @arb_baseaddr;
my @arb_highaddr;
my @arb_lasttimeslot;
my @arb_bram_init;
my @arb_bram_initp;
my $index;
my $quit;
my $odata;
my $odata_ver;
my @delay;
my $ODT_Rtt;
my $mem_num_ranks;
my $enable_ecc;
my $family;
my $mpmc3;
my $mem_type;
my @input_args;


$fout_txt = "mpmc_ctrl_path_table.txt";
$fout_ver = "mpmc_ctrl_path_params.v";
$fout_err = "ctrl_path_generation_errors.txt";
@input_args = @ARGV;    # store the input arguments 

$usage = "Usage: perl generate_ctrl_path_table.pl [-options]\n";
$usage = $usage . "          -h                  : help\n";
$usage = $usage . "          -f_txt filename.txt : text output filename\n";
$usage = $usage . "          -f_ver filename.v   : verilog output filename\n";
$usage = $usage . "          -f_err filename.txt : Error log output filename\n";
$usage = $usage . "          -c value            : clock frequency in MHz\n";
$usage = $usage . "          -tRAS value         : value of tRAS  in ns\n";
$usage = $usage . "          -tRASmax value      : value of tRASmax in ns\n";
$usage = $usage . "          -tRCD value         : value of tRCD  in ns\n";
$usage = $usage . "          -tRC value          : value of tRC   in ns\n";
$usage = $usage . "          -tRRD value         : value of tRRD  in ns\n";
$usage = $usage . "          -tWR value          : value of tWR   in ns\n";
$usage = $usage . "          -tRP value          : value of tRP   in ns\n";
$usage = $usage . "          -tMRD value         : value of tMRD  in clks\n";
$usage = $usage . "          -tRFC value         : value of tRFC  in ns\n";
$usage = $usage . "          -tWTR value         : value of tWTR  in ns\n";
$usage = $usage . "          -CAS_Latency_A_Fmax : Memory CAS Latency Max Frequency A\n";
$usage = $usage . "          -CAS_Latency_A      : Memory CAS Latency A\n";
$usage = $usage . "          -CAS_Latency_B_Fmax : Memory CAS Latency Max Frequency B\n";
$usage = $usage . "          -CAS_Latency_B      : Memory CAS Latency B\n";
$usage = $usage . "          -CAS_Latency_C_Fmax : Memory CAS Latency Max Frequency C\n";
$usage = $usage . "          -CAS_Latency_C      : Memory CAS Latency C\n";
$usage = $usage . "          -CAS_Latency_D_Fmax : Memory CAS Latency Max Frequency D\n";
$usage = $usage . "          -CAS_Latency_D      : Memory CAS Latency D\n";
$usage = $usage . "          -reg value          : 1 = memory registered, 0 = memory unbuffered\n";
$usage = $usage . "          -d value            : Size of memory data width in bits\n";
$usage = $usage . "          -part_data_width    : Data width of the part";
$usage = $usage . "          -p_cp value         : value of C_CP_PIPELINE\n";
$usage = $usage . "          -p_ap1 value        : value of C_AP_PIPELINE1\n";
$usage = $usage . "          -p_ap2 value        : value of C_AP_PIPELINE2\n";
$usage = $usage . "          -m type             : type of memory -- ddr or ddr2\n";
$usage = $usage . "          -memory_burst_length value : Memory burst length -- 2, 4, or 8, depending on memory type\n";
#$usage = $usage . "          -tStallCkeLow value      : value of first  initialization stall in ns\n";
#$usage = $usage . "          -tStallDllRst value      : value of third  initialization stall in clks\n";
$usage = $usage . "          -ArbBaseAddress index baseaddr : index decimal, baseaddr binary.\n"; 
$usage = $usage . "          -ArbLastTimeSlot index lasttimeslot : index decimal, lasttimeslot decimal.\n"; 
$usage = $usage . "          -ArbBRAMInit  index value: index decimal, value binary.\n"; 
$usage = $usage . "          -ArbBRAMInitP index value: index decimal, value binary.\n"; 

$usage = $usage . "          -tDQSS value        : value of tDQSS in clks, DDR only\n";

$usage = $usage . "          -tAL value          : value of tAL   in clks, DDR2 only\n";
$usage = $usage . "          -tRTP value         : value of tRTP  in ns, DDR2 only\n";
$usage = $usage . "          -tAOND value        : value of tAOND in clks, DDR2 only\n";
$usage = $usage . "          -tAOFD value        : value of tAOND in clks, DDR2 only\n";
$usage = $usage . "          -tCCD value         : value of tCCD  in ns, DDR2 only\n";
#$usage = $usage . "          -tStallCkeHigh value     : value of second initialization stall in ns, DDR2 only\n";
$usage = $usage . '          -mem1_odt_rtt [0/75/150/50]   : Sets value A6 and A2 in EMR1' . "\n";
$usage = $usage . "          -memory_num_ranks   : number of ranks in the memory (unused)" . "\n";
$usage = $usage . "          -enable_ecc         : 1 -> ECC; 0 -> No ECC" . "\n";
$usage = $usage . "          -family             : Family name\n";
$usage = $usage . "          -mpmc3              : If set will enable mpmc3 output" . "\n";

# DDR and DDR2
$clk_freq = -9000;
$tRAS  = -9000;
$tRASmax = -9000;
$tRCD  = -9000;
$tRC   = -9000;
$tRRD  = -9000;
$tWR   = -9000;
$tRP   = -9000;
$tMRD  = -9000;
$tRFC  = -9000;
$tWTR  = -9000;
$burst_length = -9000;
$reg_dimm = -9000;
$data_width = -9000;
$cp_pipeline = -9000;
$ap_pipeline1 = -9000;
$ap_pipeline2 = -9000;
$ddr2 = -9000;
$tStallCkeLow = -9000;
$tStallDllRst   = -9000;
$CAS_Latency_A_Fmax = -9000;
$CAS_Latency_A = -9000;
$CAS_Latency_B_Fmax = -9000;
$CAS_Latency_B = -9000;
$CAS_Latency_C_Fmax = -9000;
$CAS_Latency_C = -9000;
$CAS_Latency_D_Fmax = -9000;
$CAS_Latency_D = -9000;
$x4 = -9000;
$enable_ecc = -9000;
$family = "abcd";
$mem_type = "sdram";
#DDR only
$tDQSS = -9000;

#DDR2 only
$tAL   = -9000;
$tRTP  = -9000;
$tAOND = -9000;
$tAOFD = -9000;
$tRPA  = -9000;
$tCCD  = -9000;
$tStallCkeHigh = -9000;
$ODT_Rtt = -9000;
#$ODT_Rtt = 150;

$mpmc3 = 0;

for ($i=0;$i<16;$i++) {
    $arb_baseaddr[$i] = $i * 16;
    $arb_baseaddr[$i] = dec2bin($arb_baseaddr[$i]);
    $arb_lasttimeslot[$i] = 16;
}
for ($i=0;$i<64;$i++) {
    $arb_bram_init[$i] = 0;
}
for ($i=0;$i<8;$i++) {
    $arb_bram_initp[$i] = 0;
}

#print (join(" ", @ARGV, "\n"));    # useful to see what is passed for debugging
$ARGV = @ARGV;
for($i=0;$i<$ARGV;$i++) {
    if ($ARGV[$i] eq "-h") {
        die "$usage";
    }
    elsif ($ARGV[$i] eq "-f_txt"){
        $i++;
        $fout_txt = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-f_ver"){
        $i++;
        $fout_ver = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-f_err"){
        $i++;
        $fout_err = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-c"){
        $i++;
        $clk_freq = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tRAS"){
        $i++;
        $tRAS = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tRASmax"){
        $i++;
        $tRASmax = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tRCD"){
        $i++;
        $tRCD = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tRC"){
        $i++;
        $tRC = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tRRD"){
        $i++;
        $tRRD = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tWR"){
        $i++;
        $tWR = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tRP"){
        $i++;
        $tRP = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tMRD"){
        $i++;
        $tMRD = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tRFC"){
        $i++;
        $tRFC = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tWTR"){
        $i++;
        $tWTR = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-CAS_Latency_A_Fmax") {
        $i++;
        $CAS_Latency_A_Fmax = $ARGV[$i];
    }    
    elsif ($ARGV[$i] eq "-CAS_Latency_A") {
        $i++;
        $CAS_Latency_A = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-CAS_Latency_B_Fmax") {
        $i++;
        $CAS_Latency_B_Fmax = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-CAS_Latency_B") {
        $i++;
        $CAS_Latency_B = $ARGV[$i];
    }    
    elsif ($ARGV[$i] eq "-CAS_Latency_C_Fmax") {
        $i++;
        $CAS_Latency_C_Fmax = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-CAS_Latency_C") {
        $i++;
        $CAS_Latency_C = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-CAS_Latency_D_Fmax") {
        $i++;
        $CAS_Latency_D_Fmax = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-CAS_Latency_D") {
        $i++;
        $CAS_Latency_D = $ARGV[$i];
    }    
    elsif ($ARGV[$i] eq "-reg"){
        $i++;
        $reg_dimm = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-d"){
        $i++;
        $data_width = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-part_data_width"){
        $i++;
        if ($ARGV[$i] eq "4") {
            $x4 = 1;
        }
        else {
            $x4 = 0;
        }
    }
    elsif ($ARGV[$i] eq "-p_cp"){
        $i++;
        $cp_pipeline = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-p_ap1"){
        $i++;
        $ap_pipeline1 = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-p_ap2"){
        $i++;
        $ap_pipeline2 = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-m"){
        $i++;
        ## Kind of memory interfaced
        $mem_type = lc($ARGV[$i]);
        
        if ($mem_type eq "ddr2") {
            $ddr2 = 1;
        }
        else {
            $ddr2 = 0;
        }
    }   
    elsif ($ARGV[$i] eq "-memory_burst_length"){
        $i++;
        $burst_length = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tStallCkeLow"){
        $i++;
        $tStallCkeLow = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tStallDllRst"){
        $i++;
        $tStallDllRst = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-ArbBaseAddress") {
        $i++;
        $index = $ARGV[$i];
        $i++;
        $arb_baseaddr[$index] = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-ArbLastTimeSlot") {
        $i++;
        $index = $ARGV[$i];
        $i++;
        $arb_lasttimeslot[$index] = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-ArbBRAMInit") {
        $i++;
        $index = $ARGV[$i];
        $i++;
        $arb_bram_init[$index] = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-ArbBRAMInitP") {
        $i++;
        $index = $ARGV[$i];
        $i++;
        $arb_bram_initp[$index] = $ARGV[$i];
    }

    elsif ($ARGV[$i] eq "-tDQSS"){
        $i++;
        $tDQSS = $ARGV[$i];
    }

    elsif ($ARGV[$i] eq "-tAL"){
        $i++;
        $tAL = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tRTP"){
        $i++;
        $tRTP = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tAOND"){
        $i++;
        $tAOND = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tAOFD"){
        $i++;
        $tAOFD = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tCCD"){
        $i++;
        $tCCD = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-tStallCkeHigh"){
        $i++;
        $tStallCkeHigh = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-mem1_odt_rtt"){
        $i++;
        $ODT_Rtt = $ARGV[$i];
    }
    # dummy variable this is not used
    elsif ($ARGV[$i] eq "-memory_num_ranks"){
        $i++;
        $mem_num_ranks = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-enable_ecc"){
        $i++;
        $enable_ecc = $ARGV[$i];
    }
    elsif ($ARGV[$i] eq "-family"){
        $i++;
        $family = $ARGV[$i];
        my @family_tmp = split(//,$family);
        if (($family_tmp[0] eq 's') && ($family_tmp[1] eq 'p') && ($family_tmp[2] eq 'a') && ($family_tmp[3] eq 'r') && ($family_tmp[4] eq 't') && ($family_tmp[5] eq 'a') && ($family_tmp[6] eq 'n') && ($family_tmp[7] eq '3')) {
            $family='spartan3';
        }
    }
    elsif ($ARGV[$i] eq "-mpmc3") {
        $mpmc3 = 1; 
    }

    elsif ($ARGV[$i] eq "-UNUSED" or $ARGV[$i] eq "UNUSED"){
#        $i++;
    }

    else {
      print "Warning: ", $ARGV[$i], " is an unrecognized switch\n";
     # print ERRORFILE "Warning: ", $ARGV[$i], " is an unrecognized switch\n";    
    }
}

open(ERRORFILE, ">$fout_err") || die print "Can't open $fout_err: $!\n";
# for debugging
# print ERRORFILE (join(" ", @input_args, "\n"));    

$quit = 0;

# if we in the mpmc3 flow we will set certain parameters to be static
if ($mpmc3) {
    $ap_pipeline1 = 1;
    $ap_pipeline2 = 1;
    $cp_pipeline = 1;
}
    
if ($clk_freq == -9000)
{
  print "Error: Parameter clk_freq is missing\n";
  print ERRORFILE "Error: Parameter clk_freq is missing\n";
  $quit = 1;
}
if ($tRAS == -9000)
{
  print "Error: Parameter tRAS is missing\n";
  print ERRORFILE "Error: Parameter tRAS is missing\n";
  $quit = 1;
}
if ($tRASmax == -9000)
{
  print "Error: Parameter tRASmax is missing\n";
  print ERRORFILE "Error: Parameter tRASmax is missing\n";
  $quit = 1;
}
if ($tRCD == -9000)
{
  print "Error: Parameter tRCD is missing\n";
  print ERRORFILE "Error: Parameter tRCD is missing\n";
  $quit = 1;
}
if ($tRC == -9000)
{
  print "Error: Parameter tRC is missing\n";
  print ERRORFILE "Error: Parameter tRC is missing\n";
  $quit = 1;
}
if ($tRRD == -9000)
{
  print "Error: Parameter tRRD is missing\n";
  print ERRORFILE "Error: Parameter tRRD is missing\n";
  $quit = 1;
}
if ($tWR == -9000)
{
  print "Error: Parameter tWR is missing\n";
  print ERRORFILE "Error: Parameter tWR is missing\n";
  $quit = 1;
}
if ($tRP == -9000)
{
  print "Error: Parameter tRP is missing\n";
  print ERRORFILE "Error: Parameter tRP is missing\n";
  $quit = 1;
}
if ($tMRD == -9000)
{
  print "Error: Parameter tMRD is missing\n";
  print ERRORFILE "Error: Parameter tMRD is missing\n";
  $quit = 1;
}
if ($tRFC == -9000)
{
  print "Error: Parameter tRFC is missing\n";
  print ERRORFILE "Error: Parameter tRFC is missing\n";
  $quit = 1;
}
if ($burst_length == -9000)
{
  print "Error: Parameter burst_length is missing\n";
  print ERRORFILE "Error: Parameter memory_burst_length is missing\n";
  $quit = 1;
}
if ($reg_dimm == -9000)
{
  print "Error: Parameter reg_dimm is missing\n";
  print ERRORFILE "Error: Parameter reg_dimm is missing\n";
  $quit = 1;
}
if ($data_width == -9000)
{
  print "Error: Parameter data_width is missing\n";
  print ERRORFILE "Error: Parameter data_width is missing\n";
  $quit = 1;
}
if ($cp_pipeline == -9000)
{
  print "Error: Parameter cp_pipeline is missing\n";
  print ERRORFILE "Error: Parameter cp_pipeline is missing\n";
  $quit = 1;
}
if ($ap_pipeline1 == -9000)
{
  print "Error: Parameter ap_pipeline1 is missing\n";
  print ERRORFILE "Error: Parameter ap_pipeline1 is missing\n";
  $quit = 1;
}
if ($ap_pipeline2 == -9000)
{
  print "Error: Parameter ap_pipeline2 is missing\n";
  print ERRORFILE "Error: Parameter ap_pipeline2 is missing\n";
  $quit = 1;
}
if ($ddr2  == -9000)
{
  print "Error: Parameter ddr2 is missing\n";
  print ERRORFILE "Error: Parameter ddr2 is missing\n";
  $quit = 1;
}
#if ($tStallCkeLow == -9000)
#{
#  print "Error: Parameter tStallCkeLow is missing\n";
#  print ERRORFILE "Error: Parameter tStallCkeLow is missing\n";
#  $quit = 1;
#}
#if ($tStallDllRst == -9000)
#{
#  print "Error: Parameter tStallDllRst is missing\n";
#  print ERRORFILE "Error: Parameter tStallDllRst is missing\n";
#  $quit = 1;
#}
if ($CAS_Latency_A_Fmax == -9000)
{
  print "Error: Parameter CAS_Latency_A_Fmax is missing\n";
  print ERRORFILE "Error: Parameter CAS_Latency_A_Fmax is missing\n";
  $quit = 1;
}
if ($CAS_Latency_A == -9000)
{
  print "Error: Parameter CAS_Latency_A is missing\n";
  print ERRORFILE "Error: Parameter CAS_Latency_A is missing\n";
  $quit = 1;
}
if ($CAS_Latency_B_Fmax == -9000)
{
  print "Error: Parameter CAS_Latency_B_Fmax is missing\n";
  print ERRORFILE "Error: Parameter CAS_Latency_B_Fmax is missing\n";
  $quit = 1;
}
if ($CAS_Latency_B == -9000)
{
  print "Error: Parameter CAS_Latency_B is missing\n";
  print ERRORFILE "Error: Parameter CAS_Latency_B is missing\n";
  $quit = 1;
}
if ($CAS_Latency_C_Fmax == -9000)
{
  print "Error: Parameter CAS_Latency_C_Fmax is missing\n";
  print ERRORFILE "Error: Parameter CAS_Latency_C_Fmax is missing\n";
  $quit = 1;
}
if ($CAS_Latency_C == -9000)
{
  print "Error: Parameter CAS_Latency_C is missing\n";
  print ERRORFILE "Error: Parameter CAS_Latency_C is missing\n";
  $quit = 1;
}
if ($CAS_Latency_D_Fmax == -9000)
{
  print "Error: Parameter CAS_Latency_D_Fmax is missing\n";
  print ERRORFILE "Error: Parameter CAS_Latency_D_Fmax is missing\n";
  $quit = 1;
}
if ($CAS_Latency_D == -9000)
{
  print "Error: Parameter CAS_Latency_D is missing\n";
  print ERRORFILE "Error: Parameter CAS_Latency_D is missing\n";
  $quit = 1;
}
if ($x4 == -9000)
{
  print "Error: Parameter part_data_width is missing\n";
  print ERRORFILE "Error: Parameter part_data_width is missing\n";
  $quit = 1;
}
if ($enable_ecc == -9000)
{
  print "Error: Parameter enable_ecc is missing\n";
  print ERRORFILE "Error: Parameter enable_ecc is missing\n";
  $quit = 1;
}
if ($family eq "abcd")
{
  print "Error: Parameter family is missing\n";
  print ERRORFILE "Error: Parameter family is missing\n";
  $quit = 1;
}

if ($ddr2 == 0) {
  if ($mem_type eq "ddr")
   {
    if ($tDQSS == -9000)
    {
        print "Error: Parameter tDQSS is missing\n";
        print ERRORFILE "Error: Parameter tDQSS is missing\n";
        $quit = 1;
    }
   }
}
else {
    if ($tAL == -9000)
    {
        print "Error: Parameter tAL is missing\n";
        print ERRORFILE "Error: Parameter tAL is missing\n";
        $quit = 1;
    }
    if ($tRTP == -9000)
    {
        print "Error: Parameter tRTP is missing\n";
        print ERRORFILE "Error: Parameter tRTP is missing\n";
        $quit = 1;
    }
#    if ($tAOND == -9000)
#    {
#        print "Error: Parameter tAOND is missing\n";
#        print ERRORFILE "Error: Parameter tAOND is missing\n";
#        $quit = 1;
#    }
#    if ($tAOFD == -9000)
#    {
#        print "Error: Parameter tAOFD is missing\n";
#        print ERRORFILE "Error: Parameter tAOFD is missing\n";
#        $quit = 1;
#    }
#    if ($tStallCkeHigh == -9000)
#    {
#        print "Error: Parameter tStallCkeHigh is missing\n";
#        print ERRORFILE "Error: Parameter tStallCkeHigh is missing\n";
#        $quit = 1;
#    }
#    if ($ODT_Rtt == -9000)
#    {
#        print "Error: Parameter mem1_odt_rtt is missing\n";
#        print ERRORFILE "Error: Parameter mem1_odt_rtt is missing\n";
#        $quit = 1;
#    }
}

# Assuming these Fmax parameters will ALWAYS be in order

if($clk_freq < ($CAS_Latency_A_Fmax +1)){
  if ($CAS_Latency_A != ceil($CAS_Latency_A)) {
    print "Warning: Non integer CAS Latency $CAS_Latency_A rounded to " . ceil($CAS_Latency_A) . "\n";
    print ERRORFILE "Warning: Non integer CAS Latency $CAS_Latency_A rounded to " . ceil($CAS_Latency_A) . "\n";
  } 
  $cas_latency = ceil($CAS_Latency_A);
}
elsif($clk_freq < ($CAS_Latency_B_Fmax +1)){
  if ($CAS_Latency_B != ceil($CAS_Latency_B)) {
    print "Warning: Non integer CAS Latency $CAS_Latency_B rounded to " . ceil($CAS_Latency_B) . "\n";
    print ERRORFILE "Warning: Non integer CAS Latency $CAS_Latency_B rounded to " . ceil($CAS_Latency_B) . "\n";
  } 
  $cas_latency = ceil($CAS_Latency_B);
}  
elsif($clk_freq < ($CAS_Latency_C_Fmax +1)){
  if ($CAS_Latency_C != ceil($CAS_Latency_C)) {
    print "Warning: Non integer CAS Latency $CAS_Latency_C rounded to " . ceil($CAS_Latency_C) . "\n";
    print ERRORFILE "Warning: Non integer CAS Latency $CAS_Latency_C rounded to " . ceil($CAS_Latency_C) . "\n";
  } 
  $cas_latency = ceil($CAS_Latency_C);
}
elsif($clk_freq < ($CAS_Latency_D_Fmax +1)){
  if ($CAS_Latency_D != ceil($CAS_Latency_D)) {
    print "Warning: Non integer CAS Latency $CAS_Latency_D rounded to " . ceil($CAS_Latency_D) . "\n";
    print ERRORFILE "Warning: Non integer CAS Latency $CAS_Latency_D rounded to " . ceil($CAS_Latency_D) . "\n";
  } 
  $cas_latency = ceil($CAS_Latency_D);
}
else {
  print "Error: No suitable CAS latency found for the Clock Frequency: $clk_freq MHz\n";
  print ERRORFILE "Error: No suitable CAS latency found for the Clock Frequency: $clk_freq MHz\n";
  die "Error: No suitable CAS latency found for the Clock Frequency: $clk_freq MHz\n";
}

$odata = gensym; 
open($odata, ">$fout_txt")|| die "ERROR: $fout_txt $!\n";
print $odata "// This file has been automatically generated.  Do not modify.\n";

$odata_ver = gensym; 
open($odata_ver, ">$fout_ver")|| die "ERROR: $fout_ver $!\n";
print $odata_ver "// This file has been automatically generated.  Do not modify.\n";

$tRAS  = $tRAS  / (1000/$clk_freq);
$tRASmax = $tRASmax / (1000/$clk_freq);
$tRCD  = $tRCD  / (1000/$clk_freq);
if ($tRCD<2) {
    $tRCD = 2;
}
$tRC   = $tRC  / (1000/$clk_freq);
$tRRD  = $tRRD / (1000/$clk_freq);
$tWR_ps = $tWR * 1000;
$tWR   = $tWR   / (1000/$clk_freq);
$tRP   = $tRP   / (1000/$clk_freq);
$tMRD  = $tMRD;
$tRFC  = $tRFC  / (1000/$clk_freq);

$tDQSS = $tDQSS;

$tAL   = $tAL;
$tRTP  = $tRTP  / (1000/$clk_freq);
$tAOND = $tAOND;
$tAOFD = $tAOFD;
$tWL   = $tAL + $cas_latency - 1;
$tRPA  = $tRP + 1;

$tStallCkeLow = int $tStallCkeLow / (1000/$clk_freq);
$tStallCkeHigh = int $tStallCkeHigh / (1000/$clk_freq);
if ($ddr2==0) {
#    $tStallDllRst = $tStallDllRst - (ceil($tMRD)+ceil($tRP)+ceil($tRFC)*2+ceil($tMRD));
    $tStallDllRst = $tStallDllRst;
}
else {
#    $tStallDllRst = $tStallDllRst - (ceil($tMRD)+ceil($tRPA)+ceil($tRFC)*2+ceil($tMRD)*3);
    $tStallDllRst = $tStallDllRst;
}
if ($tStallDllRst < 0) {
    $tStallDllRst = 0;
}

for ($i=0;$i<16;$i++) {
    $arb_baseaddr[$i] =~ s/_//g ;
    $arb_baseaddr[$i] = bin2dec($arb_baseaddr[$i]);
    $arb_highaddr[$i] = $arb_baseaddr[$i] + $arb_lasttimeslot[$i] - 1;
}

if ($quit == 1) {
  die "Aborting Script: Missing arguments\n";
}

# these aren't printed in the mpmc3 flow
unless ($mpmc3) {
    # Assign Burst Length
    print $odata_ver "parameter C_BURST_LENGTH = $burst_length;\n";

    # Assign Pipeline Values
    print $odata_ver "parameter C_AP_PIPELINE1 = $ap_pipeline1;\n";

    # Assign Sequence Numbers
    print $odata_ver "parameter C_WORD_WRITE_SEQ       =  0;\n";
    print $odata_ver "parameter C_WORD_READ_SEQ        =  1;\n";
    print $odata_ver "parameter C_DOUBLEWORD_WRITE_SEQ =  2;\n";
    print $odata_ver "parameter C_DOUBLEWORD_READ_SEQ  =  3;\n";
    print $odata_ver "parameter C_CL4_WRITE_SEQ        =  4;\n";
    print $odata_ver "parameter C_CL4_READ_SEQ         =  5;\n";
    print $odata_ver "parameter C_CL8_WRITE_SEQ        =  6;\n";
    print $odata_ver "parameter C_CL8_READ_SEQ         =  7;\n";
    print $odata_ver "parameter C_B16_WRITE_SEQ        =  8;\n";
    print $odata_ver "parameter C_B16_READ_SEQ         =  9;\n";
    print $odata_ver "parameter C_B32_WRITE_SEQ        = 10;\n";
    print $odata_ver "parameter C_B32_READ_SEQ         = 11;\n";
    print $odata_ver "parameter C_B64_WRITE_SEQ        = 12;\n";
    print $odata_ver "parameter C_B64_READ_SEQ         = 13;\n";
    print $odata_ver "parameter C_REFH_SEQ             = 14;\n";
    print $odata_ver "parameter C_NOP_SEQ              = 15;\n";
}
print $odata_ver "parameter C_TWR                  = ",$tWR_ps,";\n";
# Assign Control Path Delays
print $odata_ver "parameter C_CTRL_AP_COL_DELAY = 4'h", sprintf("%01x",$ap_pipeline1), ";\n";
print $odata_ver "parameter C_CTRL_AP_PI_ADDR_CE_DELAY = 4'h", sprintf("%01x",0), ";\n";
print $odata_ver "parameter C_CTRL_AP_PORT_SELECT_DELAY = 4'h", sprintf("%01x",0), ";\n";
print $odata_ver "parameter C_CTRL_AP_PIPELINE1_CE_DELAY = 4'h", sprintf("%01x",0), ";\n";
print $odata_ver "parameter C_CTRL_DP_LOAD_RDWDADDR_DELAY = 4'h", sprintf("%01x",$ap_pipeline1 + 1), ";\n";
print $odata_ver "parameter C_CTRL_DP_RDFIFO_WHICHPORT_DELAY = 4'h", sprintf("%01x",$ap_pipeline1 + $ap_pipeline2 + $cas_latency + $tAL + 7), ";\n";
print $odata_ver "parameter C_CTRL_DP_SIZE_DELAY = 4'h", sprintf("%01x",$ap_pipeline1 + 1), ";\n";
if ($ddr2==0) {
    print $odata_ver "parameter C_CTRL_DP_WRFIFO_WHICHPORT_DELAY = 4'h", sprintf("%01x",$ap_pipeline1 + $ap_pipeline2 + $tDQSS - 1 + 1), ";\n";
}
else {
    print $odata_ver "parameter C_CTRL_DP_WRFIFO_WHICHPORT_DELAY = 4'h", sprintf("%01x",$ap_pipeline1 + $ap_pipeline2 + $tWL - 1 + 1), ";\n";
}
print $odata_ver "parameter C_CTRL_PHYIF_DUMMYREADSTART_DELAY = 4'h", sprintf("%01x",$ap_pipeline1 + $ap_pipeline2+3), ";\n";

for ($i=0;$i<36;$i++) {
    $delay[$i] = 0;
}
$delay[0] = 1;
$delay[1] = 0;
$delay[2] = $ap_pipeline1 + $ap_pipeline2;
$delay[3] = $ap_pipeline1 + $ap_pipeline2;
$delay[4] = $ap_pipeline1 + $ap_pipeline2;
$delay[5] = $ap_pipeline1 + $ap_pipeline2;
$delay[6] = $ap_pipeline1 + $ap_pipeline2 + $cas_latency;
### Move Ctrl_Skip and Ctrl_PhyIF_DQS_O forward by 1 cycle if activate takes 3 cycles or longer and not ECC.
### Move table back by 1 cycle.
if ((ceil($tRCD) - 2 > 0) && ($enable_ecc==0)) {
    $delay[7]  = 1;
    $delay[8] = $ap_pipeline1 + $ap_pipeline2 - 1;
    $delay[9]  = 1;
    $delay[11] = 1;
} else {
    $delay[7]  = 0;
    $delay[8] = $ap_pipeline1 + $ap_pipeline2 - 2;
    $delay[9]  = 0;
    $delay[11] = 0;
}
### Special Case for DDR/DDR2 Spartan3 ECC.  Move Ctrl_Skip forward by 1 cycle if Spartan3 and ECC
if (($family eq 'spartan3') && 
    (($mem_type eq "ddr") || ($mem_type eq "ddr2"))) {
    $delay[7]  = $delay[7] + 1;
    $delay[9]  = $delay[9] + 1;
    $delay[11] = $delay[11] + 1;
}
$delay[10] = $ap_pipeline1 + $ap_pipeline2 - 1;
$delay[12] = $ap_pipeline1;
$delay[13] = $ap_pipeline1;
$delay[14] = $ap_pipeline1;
$delay[15] = $ap_pipeline1;
$delay[16] = $ap_pipeline1 + $ap_pipeline2 - 1;
$delay[17] = 0;
$delay[18] = 0;
$delay[19] = 0;
$delay[20] = 0;
$delay[21] = 0;
$delay[22] = 0;
$delay[23] = 0;
$delay[24] = 0;
$delay[25] = 0;
$delay[26] = 0;
$delay[27] = 0;
$delay[28] = 0;
$delay[29] = 0;
$delay[30] = 0;
$delay[31] = 0;
if ($mem_type eq "ddr"){
    $delay[32] = $ap_pipeline1 + $ap_pipeline2 + $tDQSS + $reg_dimm - 2;
    }   
elsif ($mem_type eq "ddr2"){
    $delay[32] = $ap_pipeline1 + $ap_pipeline2 + $tWL + $reg_dimm - 2;
    }
elsif ($mem_type eq "sdram" or $mem_type eq "2"){
    $delay[32] = $ap_pipeline1 + $ap_pipeline2  - 1;
##    $delay[32] = $ap_pipeline1 + $ap_pipeline2 + $reg_dimm - 1;
    }

##if ($ddr2==0) {
##    $delay[32] = $ap_pipeline1 + $ap_pipeline2 + $tDQSS + $reg_dimm - 2;
##}
##else {
##    $delay[32] = $ap_pipeline1 + $ap_pipeline2 + $tWL + $reg_dimm - 2;
##}

#if ($mpmc3) {
#    $delay[32] = $delay[32] - 1;
#}
$delay[33] = 1;
$delay[34] = 0;
$delay[35] = 0;

print $odata_ver "parameter C_CTRL_Q0_DELAY = 4'h",  sprintf("%01x",$delay[0]), ";\n";
print $odata_ver "parameter C_CTRL_Q1_DELAY = 4'h",  sprintf("%01x",$delay[1]), ";\n";
print $odata_ver "parameter C_CTRL_Q2_DELAY = 4'h",  sprintf("%01x",$delay[2]), ";\n";
print $odata_ver "parameter C_CTRL_Q3_DELAY = 4'h",  sprintf("%01x",$delay[3]), ";\n";
print $odata_ver "parameter C_CTRL_Q4_DELAY = 4'h",  sprintf("%01x",$delay[4]), ";\n";
print $odata_ver "parameter C_CTRL_Q5_DELAY = 4'h",  sprintf("%01x",$delay[5]), ";\n";
print $odata_ver "parameter C_CTRL_Q6_DELAY = 4'h",  sprintf("%01x",$delay[6]), ";\n";
print $odata_ver "parameter C_CTRL_Q7_DELAY = 4'h",  sprintf("%01x",$delay[7]), ";\n";
print $odata_ver "parameter C_CTRL_Q8_DELAY = 4'h",  sprintf("%01x",$delay[8]), ";\n";
print $odata_ver "parameter C_CTRL_Q9_DELAY = 4'h",  sprintf("%01x",$delay[9]), ";\n";
print $odata_ver "parameter C_CTRL_Q10_DELAY = 4'h", sprintf("%01x",$delay[10]), ";\n";
print $odata_ver "parameter C_CTRL_Q11_DELAY = 4'h", sprintf("%01x",$delay[11]), ";\n";
print $odata_ver "parameter C_CTRL_Q12_DELAY = 4'h", sprintf("%01x",$delay[12]), ";\n";
print $odata_ver "parameter C_CTRL_Q13_DELAY = 4'h", sprintf("%01x",$delay[13]), ";\n";
print $odata_ver "parameter C_CTRL_Q14_DELAY = 4'h", sprintf("%01x",$delay[14]), ";\n";
print $odata_ver "parameter C_CTRL_Q15_DELAY = 4'h", sprintf("%01x",$delay[15]), ";\n";
print $odata_ver "parameter C_CTRL_Q16_DELAY = 4'h", sprintf("%01x",$delay[16]), ";\n";
print $odata_ver "parameter C_CTRL_Q17_DELAY = 4'h", sprintf("%01x",$delay[17]), ";\n";
print $odata_ver "parameter C_CTRL_Q18_DELAY = 4'h", sprintf("%01x",$delay[18]), ";\n";
print $odata_ver "parameter C_CTRL_Q19_DELAY = 4'h", sprintf("%01x",$delay[19]), ";\n";
print $odata_ver "parameter C_CTRL_Q20_DELAY = 4'h", sprintf("%01x",$delay[20]), ";\n";
print $odata_ver "parameter C_CTRL_Q21_DELAY = 4'h", sprintf("%01x",$delay[21]), ";\n";
print $odata_ver "parameter C_CTRL_Q22_DELAY = 4'h", sprintf("%01x",$delay[22]), ";\n";
print $odata_ver "parameter C_CTRL_Q23_DELAY = 4'h", sprintf("%01x",$delay[23]), ";\n";
print $odata_ver "parameter C_CTRL_Q24_DELAY = 4'h", sprintf("%01x",$delay[24]), ";\n";
print $odata_ver "parameter C_CTRL_Q25_DELAY = 4'h", sprintf("%01x",$delay[25]), ";\n";
print $odata_ver "parameter C_CTRL_Q26_DELAY = 4'h", sprintf("%01x",$delay[26]), ";\n";
print $odata_ver "parameter C_CTRL_Q27_DELAY = 4'h", sprintf("%01x",$delay[27]), ";\n";
print $odata_ver "parameter C_CTRL_Q28_DELAY = 4'h", sprintf("%01x",$delay[28]), ";\n";
print $odata_ver "parameter C_CTRL_Q29_DELAY = 4'h", sprintf("%01x",$delay[29]), ";\n";
print $odata_ver "parameter C_CTRL_Q30_DELAY = 4'h", sprintf("%01x",$delay[30]), ";\n";
print $odata_ver "parameter C_CTRL_Q31_DELAY = 4'h", sprintf("%01x",$delay[31]), ";\n";
print $odata_ver "parameter C_CTRL_Q32_DELAY = 4'h", sprintf("%01x",$delay[32]), ";\n";
print $odata_ver "parameter C_CTRL_Q33_DELAY = 4'h", sprintf("%01x",$delay[33]), ";\n";
print $odata_ver "parameter C_CTRL_Q34_DELAY = 4'h", sprintf("%01x",$delay[34]), ";\n";
print $odata_ver "parameter C_CTRL_Q35_DELAY = 4'h", sprintf("%01x",$delay[35]), ";\n";

# bram arbiter not needed in this script for mpmc3  
unless ($mpmc3) {
    # Assign Arbiter Values
    for ($i=0;$i<16;$i++) {
        print $odata_ver "parameter C_BASEADDR_ARB$i = 'h", sprintf("%03x",$arb_baseaddr[$i]), ";\n";
        print $odata_ver "parameter C_HIGHADDR_ARB$i = 'h", sprintf("%03x",$arb_highaddr[$i]), ";\n";
    }
    print $odata_ver "parameter C_ARB_BRAM_SRVAL_A = 36'b0;\n";
    print $odata_ver "parameter C_ARB_BRAM_SRVAL_B = 36'b0;\n";
    for ($i=0;$i<64;$i++) {
        print $odata_ver "parameter C_ARB_BRAM_INIT_", sprintf("%02X",$i), " = 256'b$arb_bram_init[$i];\n";
    }
    for ($i=0;$i<8;$i++) {
        print $odata_ver "parameter C_ARB_BRAM_INITP_", sprintf("%02X",$i), " = 256'b$arb_bram_initp[$i];\n";
    }
}
print $odata_ver "parameter C_USE_FIXED_BASEADDR_CTRL = 0;\n" unless ($mpmc3);

# Calculate Control Path Tables
my $start_addr = 0;
my $pattern;
my $pattern_data_width;
my $skip_0;
my $skip_1;
my $skip_2;
my $ctrl_bram_index = 0;
my @ctrl_bram_table; 
my @ctrl_bram_tablep; 
my @ctrl_bram_init;
my @ctrl_bram_initp;
my $allow_repeat;
my $repeat_value;
for ($i=0;$i<16;$i++) {
    if ($i==0) {
        $pattern = "Word Write";
        $pattern_data_width = 32;
        $allow_repeat = 0;
    }
    elsif ($i==1) {
        $pattern = "Word Read";
        $pattern_data_width = 32;
        $allow_repeat = 0;
    }
    elsif ($i==2) {
        $pattern = "DoubleWord Write";
        $pattern_data_width = 64;
        $allow_repeat = 0;
    }
    elsif ($i==3) {
        $pattern = "DoubleWord Read";
        $pattern_data_width = 64;
        $allow_repeat = 0;
    }
    elsif ($i==4) {
        $pattern = "CL4 Write";
        $pattern_data_width = 32 * 4;
        $allow_repeat = 0;
    }
    elsif ($i==5) {
        $pattern = "CL4 Read";
        $pattern_data_width = 32 * 4;
        $allow_repeat = 0;
    }
    elsif ($i==6) {
        $pattern = "CL8 Write";
        $pattern_data_width = 32 * 8;
        $allow_repeat = 0;
    }
    elsif ($i==7) {
        $pattern = "CL8 Read";
        $pattern_data_width = 32 * 8;
        $allow_repeat = 0;
    }
    elsif ($i==8) {
        $pattern = "B16 Write";
        $pattern_data_width = 32 * 16;
        if ($enable_ecc == 1 && $mem_type eq "sdram") {
            $allow_repeat = 1;
            $repeat_value = $pattern_data_width/($data_width)/4-2;
            print $odata_ver "parameter C_B16_REPEAT_CNT = $repeat_value;\n";

        }
        else {
            print $odata_ver "parameter C_B16_REPEAT_CNT = 0;\n";
            $allow_repeat = 0;
        }
    }
    elsif ($i==9) {
        $pattern = "B16 Read";
        $pattern_data_width = 32 * 16;
        if ($enable_ecc == 1 && $mem_type eq "sdram") {
            $allow_repeat = 1;
        }
        else {
            $allow_repeat = 0;
        }
    }
    elsif ($i==10) {
        $pattern = "B32 Write";
        $pattern_data_width = 32 * 32;
        $allow_repeat = 1;
        if ($mem_type eq "sdram") {
            $repeat_value = $pattern_data_width/($data_width)/4-2;
        }
        else {
            $repeat_value = $pattern_data_width/($data_width*2)/4-2;
        }
        print $odata_ver "parameter C_B32_REPEAT_CNT = $repeat_value;\n";
    }
    elsif ($i==11) {
        $pattern = "B32 Read";
        $pattern_data_width = 32 * 32;
        $allow_repeat = 1;
    }
    elsif ($i==12) {
        $pattern = "B64 Write";
        $pattern_data_width = 32 * 64;
        $allow_repeat = 1;
        if ($mem_type eq "sdram") {
            $repeat_value = $pattern_data_width/($data_width)/4-2;
        }
        else {
            $repeat_value = $pattern_data_width/($data_width*2)/4-2;
        }
        print $odata_ver "parameter C_B64_REPEAT_CNT = $repeat_value;\n";
    }
    elsif ($i==13) {
        $pattern = "B64 Read";
        $pattern_data_width = 32 * 64;
        $allow_repeat = 1;
    }
    elsif ($i==14) {
        $pattern = "Refresh";
        $pattern_data_width = 0;
        $allow_repeat = 0;
    }
    elsif ($i==15) {
        $pattern = "NOP";
        $pattern_data_width = 0;
        $allow_repeat = 0;
    }
    if ($allow_repeat==1) {
        $num_data_beats=8;
    } else {
        $num_data_beats = 0;
        if ($mem_type eq "sdram")
        {
            while ($num_data_beats * $data_width  < $pattern_data_width) {
                $num_data_beats++;    
            }
        }
        else
        {
            while ($num_data_beats * $data_width * 2 < $pattern_data_width) {
                $num_data_beats++;
            }
        }
    }
      
    if ($mem_type eq "ddr") {
        if ((($i % 2) == 0) && ($i < 14)) {
            if ($enable_ecc == 1) {
                if ($i==0) {
                    $skip_2 = 0; $skip_1 = 0; $skip_0 = 1; 
                }
                elsif ($i==2) {
                    $skip_2 = 0; $skip_1 = 1; $skip_0 = 0; 
                }
                elsif ($i==4) {
                    $skip_2 = 0; $skip_1 = 1; $skip_0 = 1; 
                }
                elsif ($i==6) {
                    $skip_2 = 1; $skip_1 = 0; $skip_0 = 0; 
                }
                elsif ($i==8) {
                    $skip_2 = 1; $skip_1 = 0; $skip_0 = 1; 
                }
                elsif ($i==10) {
                    $skip_2 = 1; $skip_1 = 1; $skip_0 = 0; 
                }
                elsif ($i==12) {
                    $skip_2 = 1; $skip_1 = 1; $skip_0 = 1; 
                }
                @table = &ddr_activate_to_read_to_write_to_precharge_to_activate($tRCD,$tDQSS,$num_data_beats,$tWR,$burst_length,$tRAS,$tRASmax,$tRP,$tRC,$tRRD,$cas_latency,$skip_0,$skip_1,$skip_2,$allow_repeat,$family);
            }
            else {
                @table = &ddr_activate_to_write_to_precharge_to_activate($tRCD,$tDQSS,$num_data_beats,$tWR,$burst_length,$tRAS,$tRASmax,$tRP,$tRC,$tRRD,$allow_repeat,$family);
            }
            # Add extra cycle to write table to account for delayed
            # RAS/CAS/WE.  In future, look at adding stall to arbiter
            # in place of this work around.
            my @table_tmp = &ddr_nop(1);
            for ($j=0;$j<36;$j++) {
                $table[$j] = $table[$j] . $table_tmp[$j];
            }
            my @table_0 = split(//,$table[0]);
            my $table_0_length = length($table[0]);
            $table[0] = '0';
            for($j=0;$j<$table_0_length-1;$j++) {
                $table[0] = $table[0] . $table_0[$j];
            }

            my $table_16_length = length($table[16]);
            $table[16] = '';
            for($j=0;$j<$table_16_length;$j++) {
                $table[16] = $table[16] . '1';
            }
        }
        elsif ((($i % 2) == 1) && ($i < 14)) {
            @table = &ddr_activate_to_read_to_precharge_to_activate($tRCD,$num_data_beats,$burst_length,$tRAS,$tRASmax,$tRP,$tRC,$tRRD,$allow_repeat,$family);
        }
        elsif ($i == 14) {
            @table = &ddr_precharge_to_refresh_to_activate($tRP,$tRFC,$family);
        }
        elsif ($i == 15) {
            @table = &ddr_nop(2);
        }
    }
    elsif ($mem_type eq "ddr2") {
        if ((($i % 2) == 0) && ($i < 14)) {
            if ($enable_ecc == 1) {
                if ($i==0) {
                    $skip_2 = 0; $skip_1 = 0; $skip_0 = 1; 
                }
                elsif ($i==2) {
                    $skip_2 = 0; $skip_1 = 1; $skip_0 = 0; 
                }
                elsif ($i==4) {
                    $skip_2 = 0; $skip_1 = 1; $skip_0 = 1; 
                }
                elsif ($i==6) {
                    $skip_2 = 1; $skip_1 = 0; $skip_0 = 0; 
                }
                elsif ($i==8) {
                    $skip_2 = 1; $skip_1 = 0; $skip_0 = 1; 
                }
                elsif ($i==10) {
                    $skip_2 = 1; $skip_1 = 1; $skip_0 = 0; 
                }
                elsif ($i==12) {
                    $skip_2 = 1; $skip_1 = 1; $skip_0 = 1; 
                }
                @table = &ddr2_activate_to_read_to_write_to_precharge_to_activate($tRCD,$tAL,$num_data_beats,$burst_length,$tRTP,$tRAS,$tRASmax,$tRPA,$tRC,$tRRD,$tWL,$tWR,$ODT_Rtt,$cas_latency,$skip_0,$skip_1,$skip_2,$allow_repeat,$family);
            }
            else {
                @table = &ddr2_activate_to_write_to_precharge_to_activate($tRCD,$tWL,$num_data_beats,$tWR,$burst_length,$tRAS,$tRASmax,$tRPA,$tRC,$tRRD,$ODT_Rtt,$allow_repeat,$family);
            }
            # Add extra cycle to write table to account for delayed
            # RAS/CAS/WE.  In future, look at adding stall to arbiter
            # in place of this work around.
            my@table_tmp = &ddr2_nop(1);
            for ($j=0;$j<36;$j++) {
                $table[$j] = $table[$j] . $table_tmp[$j];
            }
            my @table_0 = split(//,$table[0]);
            my $table_0_length = length($table[0]);
            $table[0] = '0';
            for($j=0;$j<$table_0_length-1;$j++) {
                $table[0] = $table[0] . $table_0[$j];
            }

            my $table_16_length = length($table[16]);
            $table[16] = '';
            for($j=0;$j<$table_16_length;$j++) {
                $table[16] = $table[16] . '1';
            }
        }
        elsif ((($i % 2) == 1) && ($i < 14)) {
            @table = &ddr2_activate_to_read_to_precharge_to_activate($tRCD,$tAL,$num_data_beats,$burst_length,$tRTP,$tRAS,$tRASmax,$tRPA,$tRC,$tRRD,$allow_repeat,$family);
        }
        elsif ($i == 14) {
            @table = &ddr2_precharge_to_refresh_to_activate($tRPA,$tRFC,$family);
        }
        elsif ($i == 15) {
            @table = &ddr2_nop(2);
        }
    }
    elsif ($mem_type eq "sdram") {
        if ((($i % 2) == 0) && ($i < 14)) {
            if ($enable_ecc == 1) {
                if ($i==0) {
                    $skip_2 = 0; $skip_1 = 0; $skip_0 = 1; 
                }
                elsif ($i==2) {
                    $skip_2 = 0; $skip_1 = 1; $skip_0 = 0; 
                }
                elsif ($i==4) {
                    $skip_2 = 0; $skip_1 = 1; $skip_0 = 1; 
                }
                elsif ($i==6) {
                    $skip_2 = 1; $skip_1 = 0; $skip_0 = 0; 
                }
                elsif ($i==8) {
                    $skip_2 = 1; $skip_1 = 0; $skip_0 = 1; 
                }
                elsif ($i==10) {
                    $skip_2 = 1; $skip_1 = 1; $skip_0 = 0; 
                }
                elsif ($i==12) {
                    $skip_2 = 1; $skip_1 = 1; $skip_0 = 1; 
                }
                @table = &sdr_activate_to_read_to_write_to_precharge_to_activate($tRCD,0,$num_data_beats,$tWR,$burst_length,$tRAS,$tRASmax,$tRP,$tRC,$tRRD,$cas_latency,$skip_0,$skip_1,$skip_2,$allow_repeat,$family);
            }
            else {
                @table = &sdr_activate_to_write_to_precharge_to_activate($tRCD,0,$num_data_beats,$tWR,$burst_length,$tRAS,$tRASmax,$tRP,$tRC,$tRRD,$allow_repeat,$family);
            }
            # Add extra cycle to write table to account for delayed
            # RAS/CAS/WE.  In future, look at adding stall to arbiter
            # in place of this work around.
            my @table_tmp = &sdr_nop(1);
            for ($j=0;$j<36;$j++) {
                $table[$j] = $table[$j] . $table_tmp[$j];
            }
            my @table_0 = split(//,$table[0]);
            my $table_0_length = length($table[0]);
            $table[0] = '0';
            for($j=0;$j<$table_0_length-1;$j++) {
                $table[0] = $table[0] . $table_0[$j];
            }

            my $table_16_length = length($table[16]);
            $table[16] = '';
            for($j=0;$j<$table_16_length;$j++) {
                $table[16] = $table[16] . '1';
            }
        }
        elsif ((($i % 2) == 1) && ($i < 14)) {
            @table = &sdr_activate_to_read_to_precharge_to_activate($tRCD,$num_data_beats,$burst_length,$tRAS,$tRASmax,$tRP,$tRC,$tRRD,$allow_repeat,$family);
        }
        elsif ($i == 14) {
            @table = &sdr_precharge_to_refresh_to_activate($tRP,$tRFC,$family);
        }
        elsif ($i == 15) {
            @table = &sdr_nop(2);
        }
    }
    

    # here we will check that tCCD is not violated for DDR2
    if ($ddr2 == 1) { 
    for($j=($tCCD-1); $j<length($table[3]) ; $j++) {
    if (substr($table[3], $j, 1) == 0) {
    foreach $k (1 .. $tCCD-1) {
    if (substr($table[3], $j-$k, 1) == 0) {
        print "Error: tCCD value of $tCCD has been violated. For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
        print ERRORFILE "Error: tCCD value of $tCCD has been violated. For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
        die "Error: tCCD value of $tCCD has been violated. For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n"; 
    }
    }
    }
    }
    }

    ### Move Ctrl_Skip and Ctrl_PhyIF_DQS_O back by 1 cycle if activate takes 3 cycles or longer and not ECC.
    ### Move delay forward by 1 cycle.
    if ((ceil($tRCD) - 2 > 0) && ($enable_ecc==0)) {
        my @table_7 = split(//,$table[7]);
        my $table_7_length = length($table[7]);
        $table[7] = '';
        for($j=1;$j<$table_7_length;$j++) {
            $table[7] = $table[7] . $table_7[$j];
        }
        $table[7] = $table[7] . '0';

        my @table_8 = split(//,$table[8]);
        my $table_8_length = length($table[8]);
        $table[8] = '';
        for($j=1;$j<$table_8_length;$j++) {
            $table[8] = $table[8] . $table_8[$j];
        }
        $table[8] = $table[8] . '0';

        my @table_9 = split(//,$table[9]);
        my $table_9_length = length($table[9]);
        $table[9] = '';
        for($j=1;$j<$table_9_length;$j++) {
            $table[9] = $table[9] . $table_9[$j];
        }
        $table[9] = $table[9] . '0';

        my @table_11 = split(//,$table[11]);
        my $table_11_length = length($table[11]);
        $table[11] = '';
        for($j=1;$j<$table_11_length;$j++) {
            $table[11] = $table[11] . $table_11[$j];
        }
        $table[11] = $table[11] . '0';
    }

    ### Move Ctrl_DP_RdFIFO_Push back by 1 cycle if not ECC.
    if ($enable_ecc == 0) {
        my @table_10 = split(//,$table[10]);
        my $table_10_length = length($table[10]);
        $table[10] = '';
        for($j=1;$j<$table_10_length;$j++) {
            $table[10] = $table[10] . $table_10[$j];
        }
        $table[10] = $table[10] . '0';
    }

    ### Move Ctrl_AP_Col_Cnt_Load back by 1 cycle.
    my @table_12 = split(//,$table[12]);
    my $table_12_length = length($table[12]);
    $table[12] = '';
    for($j=1;$j<$table_12_length;$j++) {
        $table[12] = $table[12] . $table_12[$j];
    }
    $table[12] = $table[12] . '0';
    
    ### Move Ctrl_AP_Col_Cnt_Enable back by 1 cycle.
    my @table_13 = split(//,$table[13]);
    my $table_13_length = length($table[13]);
    $table[13] = '';
    for($j=1;$j<$table_13_length;$j++) {
        $table[13] = $table[13] . $table_13[$j];
    }
    $table[13] = $table[13] . '0';
    

    ### If ECC, move Ctrl_PhyIF_DQS_O back by 2 cycles, otherwise move back
    ### by 1 cycle.
    if ($enable_ecc == 1) {
        my @table_8 = split(//,$table[8]);
        my $table_8_length = length($table[8]);
        $table[8] = '';
        for($j=2;$j<$table_8_length;$j++) {
            $table[8] = $table[8] . $table_8[$j];
        }
        $table[8] = $table[8] . '00';   
    }
    else {
        my @table_8 = split(//,$table[8]);
        my $table_8_length = length($table[8]);
        $table[8] = '';
        for($j=1;$j<$table_8_length;$j++) {
            $table[8] = $table[8] . $table_8[$j];
        }
        $table[8] = $table[8] . '0';   
    }
    if ($enable_ecc == 1) {
        ### If ECC move Ctrl_PhyIF_Force_DM back by 3 cycles.
        my @table_32 = split(//,$table[32]);
        my $table_32_length = length($table[32]);
        $table[32] = '';
        for($j=3;$j<$table_32_length;$j++) {
            $table[32] = $table[32] . $table_32[$j];
        }

        # check if table is greater than 2, then move back 3 spots, else move back 2 (should never be 1)
        if ($table_32_length > 2) {
            $table[32] = $table[32] . '000';
        }
        else {
            $table[32] = $table[32] . '00';
        }
             
    }
    else {
        ### If not ECC move Ctrl_PhyIF_Force_DM back by 2 cycle.
        my @table_32 = split(//,$table[32]);
        my $table_32_length = length($table[32]);
        $table[32] = '';
        for($j=2;$j<$table_32_length;$j++) {
            $table[32] = $table[32] . $table_32[$j];
        }
        $table[32] = $table[32] . '00';
    }

    ### If SDRAM, move Ctrl_PhyIF_Force_DM back by 1 more cycle.
    ### This will not work with ECC.
    if ($mem_type eq "sdram") {
        my @table_32 = split(//,$table[32]);
        my $table_32_length = length($table[32]);
	while ($table_32[0] != 0) {
	    @table_tmp = &sdr_nop(1);
	    for ($j=0;$j<36;$j++) {
		$table[$j] = $table_tmp[$j] . $table[$j];
	    }
	    @table_32 = split(//,$table[32]);
	    $table_32_length = length($table[32]);
	}
        $table[32] = '';
        for($j=1;$j<$table_32_length;$j++) {
            $table[32] = $table[32] . $table_32[$j];
        }
        $table[32] = $table[32] . '0';
    }

    ### Move Ctrl_Repeat4 back by 3 cycles.
    my @table_33 = split(//,$table[33]);
    my $table_33_length = length($table[33]);
    $table[33] = '';
    if ($enable_ecc == 1 && $mem_type eq "sdram") {
        for($j=4;$j<$table_33_length;$j++) {
            $table[33] = $table[33] . $table_33[$j];
        }
        # check if table is greater than 2, then move back 4 spots, 
        # else move back 2 (should never be 1)
        if ($table_33_length > 2) {
            $table[33] = $table[33] . '0000';
        }
        else {
            $table[33] = $table[33] . '00';
        }
    }
    else {
        for($j=3;$j<$table_33_length;$j++) {
            $table[33] = $table[33] . $table_33[$j];
         }    
        # check if table is greater than 2, then move back 3 spots, 
        # else move back 2 (should never be 1)
         
        if ($table_33_length > 2) {
            $table[33] = $table[33] . '000';
        }
        else {
            $table[33] = $table[33] . '00';
        }
    }

    # Check table lengths
    my $tmp = length($table[0]);
    for ($j=0;$j<36;$j++) {
        if ($tmp != length($table[$j])) {
            print "Error: $pattern table not generated correctly.  Inconsistent lengths.  For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
            for (my $k=0;$k<36;$k++) {
                print "$table[$k]\n";
            }
            print ERRORFILE "Error: $pattern table not generated correctly.  Inconsistent lengths.\n";
            for (my $k=0;$k<36;$k++) {
                print ERRORFILE "$table[$k]\n";
            }
            die "Aborting Script: $pattern table not generated correctly.  Inconsistent lengths.  For more information on this error, please consult the Answers Database or open a WebCase with this project attached at http://www.xilinx.com/support.\n";
        }
    }
    
    my $end_addr = $start_addr + length($table[0]) - 1;
    print $odata_ver "parameter C_BASEADDR_CTRL$i = 9'h", sprintf("%03x",$start_addr), ";\n";
    print $odata_ver "parameter C_HIGHADDR_CTRL$i = 9'h", sprintf("%03x",$end_addr), ";\n";
    $start_addr = $end_addr + 1;
    if ($end_addr > 2**9-1) {
        print "Error: C_HIGHADDR_CTRL$i too large.  Need to reduce sequence lengths.\n";
        print ERRORFILE "Error: C_HIGHADDR_CTRL$i too large.  Need to reduce sequence lengths.\n";
        die "Aborting Script: C_HIGHADDR_CTRL$i too large.  Need to reduce sequence lengths.\n";
    }

    ### Calculate length of skip
    my $skip_value;
    if ($enable_ecc == 1) {
        my $table_2_length = length($table[2]);
        my @table_2 = split(//,$table[2]);
        my @table_3 = split(//,$table[3]);
        my @table_4 = split(//,$table[4]);
        my $read_index = 0;
        my $write_index = 0;
        for($j=0;$j<$table_2_length;$j++) {
            if (($table_2[$j]==1) && ($table_3[$j]==0) && ($table_4[$j]==1)) {
                $read_index = $j;
            }
            if (($table_2[$j]==1) && ($table_3[$j]==0) && ($table_4[$j]==0)) {
                $write_index = $j;
            }
        }
        $skip_value = $write_index - $read_index + 1;
    }
    else {
        $skip_value = 1;
    }
    $skip_value=sprintf("%03x",$skip_value);
    if ($i==0) {
        print $odata_ver "parameter integer C_SKIP_1_VALUE = 9'h$skip_value;\n";
    }
    elsif ($i==2) {
        print $odata_ver "parameter integer C_SKIP_2_VALUE = 9'h$skip_value;\n";
    }
    elsif ($i==4) {
        print $odata_ver "parameter integer C_SKIP_3_VALUE = 9'h$skip_value;\n";
    }
    elsif ($i==6) {
        print $odata_ver "parameter integer C_SKIP_4_VALUE = 9'h$skip_value;\n";
    }
    elsif ($i==8) {
        print $odata_ver "parameter integer C_SKIP_5_VALUE = 9'h$skip_value;\n";
    }
    elsif ($i==10) {
        print $odata_ver "parameter integer C_SKIP_6_VALUE = 9'h$skip_value;\n";
    }
    elsif ($i==12) {
        print $odata_ver "parameter integer C_SKIP_7_VALUE = 9'h$skip_value;\n";
    }
    
    print $odata "//--------------------------------------------------------------------------\n";
    print $odata "//\n";
    print $odata "// FSM PATTERN $i: $pattern\n";
    print $odata "//\n";
    print $odata "// Control Signals                     0         1         2         3\n";
    print $odata "// (32 Signals)                        01234567890123456789012345678901  Comments\n";
    print $odata "// ---------------                     --------------------------------  ------------------------------\n";   
    print $odata "/*     0 Ctrl_Complete                */  $table[0]  // Delayed by $delay[0]\n";
    print $odata "/*     1 Ctrl_Stall                   */  $table[1]  // Delayed by $delay[1]\n";
    print $odata "/*     2 Ctrl_PhyIF_RAS_n             */  $table[2]  // Delayed by $delay[2]\n";
    print $odata "/*     3 Ctrl_PhyIF_CAS_n             */  $table[3]  // Delayed by $delay[3]\n";
    print $odata "/*     4 Ctrl_PhyIF_WE_n              */  $table[4]  // Delayed by $delay[4]\n";
    print $odata "/*     5 Ctrl_PhyIF_CE                */  $table[5]  // Delayed by $delay[5]\n";
    print $odata "/*     6 Ctrl_RMW                     */  $table[6]  // Delayed by $delay[6]\n";
    print $odata "/*     7 Ctrl_Skip[0]                 */  $table[7]  // Delayed by $delay[7]\n";
    print $odata "/*     8 Ctrl_PhyIF_DQS_O             */  $table[8]  // Delayed by $delay[8]\n";
    print $odata "/*     9 Ctrl_Skip[1]                 */  $table[9]  // Delayed by $delay[9]\n";
    print $odata "/*    10 Ctrl_DP_RdFIFO_Push          */  $table[10]  // Delayed by $delay[10]\n";
    print $odata "/*    11 Ctrl_Skip[2]                 */  $table[11]  // Delayed by $delay[11]\n";
    print $odata "/*    12 Ctrl_AP_Col_Cnt_Load         */  $table[12]  // Delayed by $delay[12]\n";
    print $odata "/*    13 Ctrl_AP_Col_Cnt_Enable       */  $table[13]  // Delayed by $delay[13]\n";
    print $odata "/*    14 Ctrl_AP_Precharge_Addr10     */  $table[14]  // Delayed by $delay[14]\n";
    print $odata "/*    15 Ctrl_AP_Row_Col_Sel          */  $table[15]  // Delayed by $delay[15]\n";
    print $odata "/*    16 Ctrl_Is_Write                */  $table[16]  // Delayed by $delay[16]\n";
    for ($k=17;$k<=29;$k++) {
        print $odata "/*    $k Unused                       */  $table[$k]  // Delayed by $delay[$k]\n";
    }
    for ($k=30;$k<=31;$k++) {
        $tmp = $k-30;
        print $odata "/*    $k Unused                       */  $table[$k]  // Delayed by $delay[$k]\n";
    }
    print $odata "/*    32 Ctrl_PhyIF_Force_DM          */  $table[32]  // Delayed by $delay[32]\n";
    print $odata "/*    33 Ctrl_Repeat4                 */  $table[33]  // Delayed by $delay[33]\n";
    for ($k=34;$k<=35;$k++) {
        print $odata "/*    $k Unused                       */  $table[$k]  // Delayed by $delay[$k]\n";
    }

    if ($mpmc3) {
        for (my $col = 0; $col < length($table[0]); $col++) {
            for (my $table_row = 0; $table_row < 32; $table_row++) {
                unshift (@ctrl_bram_table, (substr($table[$table_row],$col,1)));
            }
            for (my $table_row = 32; $table_row < 36; $table_row++) {
                unshift (@ctrl_bram_tablep, (substr($table[$table_row],$col,1)));
            }
            $ctrl_bram_index = $ctrl_bram_index + 1;
            if ($ctrl_bram_index % 8 == 0) {
                my $hex_table =  unpack("H*", pack("B*", join("",@ctrl_bram_table)));
                $ctrl_bram_init[($ctrl_bram_index/8)-1] = $hex_table;
                #print $odata_ver "parameter C_CTRL_BRAM_INIT_", sprintf("%02X",($ctrl_bram_index/8)-1), " = 256'h", $hex_table, ";\n";
                #print $odata_ver "parameter C_CTRL_BRAM_INIT_", sprintf("%02X",($ctrl_bram_index/8)-1), " = 256'h", $hex_table, ";\n";
                @ctrl_bram_table = ();
            }
            if ($ctrl_bram_index % 64 == 0) {
                my $hex_table =  unpack("H*", pack("B*", join("",@ctrl_bram_tablep)));
                $ctrl_bram_initp[($ctrl_bram_index/64)-1] = $hex_table;
                #print $odata_ver "parameter C_CTRL_BRAM_INIT_", sprintf("%02X",($ctrl_bram_index/8)-1), " = 256'h", $hex_table, ";\n";
                @ctrl_bram_tablep = ();
            }
        }
    }
        
        

}
if ($mpmc3) {
    # let's pad the rest with no ops
    while ($ctrl_bram_index <= 0x40*8) {
        unshift(@ctrl_bram_table, "00000000000000000000001011111100");  # pad with noop commands
        unshift(@ctrl_bram_tablep, "0000");
        $ctrl_bram_index = $ctrl_bram_index + 1;
        if ($ctrl_bram_index % 8 == 0) {
            my $hex_table =  unpack("H*", pack("B*", join("",@ctrl_bram_table)));
            $ctrl_bram_init[($ctrl_bram_index/8)-1] = $hex_table;
            @ctrl_bram_table = ();
        }
        if ($ctrl_bram_index % 64 == 0) {
            my $hex_table =  unpack("H*", pack("B*", join("",@ctrl_bram_tablep)));
            $ctrl_bram_initp[($ctrl_bram_index/64)-1] = $hex_table;
            @ctrl_bram_tablep = ();
        }
    }
   
    for (my $i = 0x3F ; $i >= 0; $i-- ) {
        print $odata_ver "parameter C_CTRL_BRAM_INIT_", sprintf("%02X",$i), " = 256'h", uc($ctrl_bram_init[$i]), ";\n";
    }
#    print $odata_ver "parameter C_CTRL_BRAM_INIT = 36'h0000002FC;\n";
    print $odata_ver "parameter C_CTRL_BRAM_SRVAL = 36'h0000002FC;\n";
    for (my $i = 0x7 ; $i >= 0; $i-- ) {
        print $odata_ver "parameter C_CTRL_BRAM_INITP_", sprintf("%02X",$i), " = 256'h", uc($ctrl_bram_initp[$i]), ";\n";
    }

    # output the cas latency while we are at it
    print $odata_ver "parameter C_MEM_CAS_LATENCY0 = 4'd" . $cas_latency . ";\n";
}
    
    


close(ERRORFILE);
close ($odata);
close ($odata_ver);
